import yaml
import numpy as np


class coarseningInit(object):
    def __init__(self, empty=False, file_path=None):
        if file_path is None:
            file_path = 'input_cards/coarsening.yml'
        if empty is False:
            self.read(file_path)
        else:
            self.tree = {'Scheme': '', 'Simple': '', 'Smart': ''}

    def scheme(self, scheme='smart'):
        self.tree['Scheme'] = scheme

    def simple(self, nx=4, ny=4, nz=4):
        self.tree['Scheme'] = 'simple'
        self.tree['Simple'] = {'nx': nx, 'ny': ny, 'nz': nz}

    def smart(self, path='default', file='mesh.h5m'):
        self.tree['Scheme'] = 'smart'
        self.tree['Smart'] = {'path': path, 'file': file}

    def read(self, file_path):
        with open(file_path) as file:  # Use file to refer to the file object
            data = file.read()
            self.tree = yaml.safe_load(data)

    # def write(self, file_path):
    #     with open(file_path, 'w') as f:
    #         data = yaml.dump(self.tree, f)


class variableInit(object):
    def __init__(self, empty=False, file_path=None):
        self.level_tree = {}
        if file_path is None:
            file_path = 'input_cards/variable_input.yml'
        if empty is False:
            self.read(file_path)
        else:
            self.tree = np.array([[], [], [], [], [], [], []]).T

    def get_var(self, level=0, coarse_num=0):
        ref = self.tree[:, 0].astype('int') == level
        level_entries = self.tree[ref]
        # self.tree = np.delete(self.tree, np.where(ref)[0], 0)
        return self.create_command(level_entries, level, coarse_num)

    def create_command(self, var_list, level, coarse_num):
        command_list = []
        cmd_pattern = 'self.{0} = MoabVariable(self.core, data_size={1}, var_type=\'{2}\', data_format=\'{3}\', name_tag=\'{0}\', data_density=\'{4}\')'
        cmd_pattern_level = 'self.{0} = MoabVariableMS(self.core, data_size={1}, var_type=\'{2}\',level={5}, coarse_num={6}, data_format=\'{3}\', name_tag=\'{0}\', data_density=\'{4}\')'
        density = ['dense', 'sparse']
        if level == 0:
            for var in var_list:
                sparse = density[bool(var[-2])]
                command_list.append(cmd_pattern.format(var[1], var[3], var[2], var[-3], sparse))
        else:
            for var in var_list:
                sparse = density[bool(var[-2])]
                command_list.append(cmd_pattern_level.format(var[1], var[3], var[2], var[4], sparse, level, coarse_num))
        return command_list

    def add_entry(self, name, type, data_size, data_type, sparse=False,
                  level=0):
        entry = [level, name, type, data_size, data_type, sparse, []]
        self.tree = np.vstack((self.tree, np.array(entry)))

    def read(self, file_path):
        with open(file_path) as file:
            data = file.read()
            tree = yaml.safe_load(data)
        self.tree = self.list_variable(tree)

    def list_variable(self, tree):
        all_var = []
        for var in list(tree):
            all_var.append([int(tree[var]['level']), var,
                            tree[var]['type'],
                            int(tree[var]['data size']),
                            tree[var]['data type'], bool(tree[var]['sparse']),
                            list()])
        return np.array(all_var)




# Run preprocessor

# docker run -t -it -v  /home/arturcastiel/projetos:/pytest desenvolvimento:latest bash -c "cd /pytest; bash"
# docker run -t -it -v  /home/arturcastiel/projetos:/pytest gabrielmmats/impress bash -c "cd /pytest; bash"
# docker run -it -v  /home/arturcastiel/projetos:/pytest desenvolvimento:latest bash -c "cd /pytest; bash"
# %load_ext autoreload
# %autoreload 2
import numpy as np
import time
from preprocessor.meshHandle.multiscaleMesh import FineScaleMeshMS as msh
from preprocessor.meshHandle.dualCoarseMesh import DualCoarseMesh as dual
import preprocessor.geoUtil.geoTools as gtool


start = time.time()
M = msh('mesh/20.h5m', dim = 3)
# M = msh('mesh/malha03.msh', dim = 2)
end = time.time()
print("The preprocessing step lasted {0}s".format(end-start))
#l = dual(M)



import numpy as np
from preprocessor.meshHandle.configTools.configClass import coarseningInit as cm
from preprocessor.msCoarseningLib.partitionTools import partitionManager as pm
# from .preprocessor.meshHandle.finescaleMesh import FineScaleMesh

def plane_check(coords_list, elements, normal_plane, point_on_plane):


    pass


config_object = cm(empty=True)
config_object.smart(file='semi.msh')
former = pm(M, config_object)
former.run()



coord_list = M.nodes.coords[:]
elements = M.volumes.connectivities[:]
normal_plane = former.partitioner.dual.faces.normal[0]
point_on_plane = former.partitioner.dual.faces.center[0]



"""This is the mpfad-3D preprocessor Tests."""
import unittest
import numpy as np
from mesh_preprocessor import Mesh


class PreMpfaDTest(unittest.TestCase):

    def setUp(self):
        K_1 = [1.0, 0.0, 0.0,
               0.0, 1.0, 0.0,
               0.0, 0.0, 1.0]
        self.mesh = Mesh('mesh/test_mesh_5_vols.h5m', dim=3)
        self.mesh.set_boundary_conditions('Dirichlet', {101: 0.0})
        self.mesh.set_material_prop('permeability', {1: K_1})

    def tearDown(self):
        del self.mesh

    def test_preprocessor_class_should_be_none(self):
        """Test class initiatilization."""
        mesh = self.mesh
        self.assertIsNotNone(mesh)

    # def test_get_all_volumes(self):
    #     """Test class get all volumes."""
    #     volumes = self.mesh.all_volumes
    #     self.assertEqual(len(volumes), 9)
    #
    # def test_get_all_internal_faces(self):
    #     """Test class get all internal faces."""
    #     in_faces = self.mesh.in_faces
    #     self.assertEqual(len(in_faces), 12)
    #
    # def test_get_all_boundary_faces(self):
    #     """Test class get all boundary faces."""
    #     b_faces = self.mesh.b_faces
    #     self.assertEqual(len(b_faces), 16)
    #
    # def test_get_all_internal_verts(self):
    #     """Test class get all internal verts."""
    #     in_verts = self.mesh.in_verts
    #     self.assertEqual(len(in_verts), 0)
    #
    # def test_get_all_boundary_verts(self):
    #     """Test class get all boundary verts."""
    #     b_verts = self.mesh.b_verts
    #     self.assertEqual(len(b_verts), 13)
    #
    # def test_screening_faces_gets_only_quadrilateral_intern_faces(self):
    #     """Test class to get only quadrilateral faces."""
    #     in_faces = self.mesh.in_faces
    #     _, in_quad_faces = self.mesh.screen_faces_by_verts(in_faces)
    #     self.assertEqual(len(in_quad_faces),  1)
    #
    # def test_screening_faces_gets_only_triangular_intern_faces(self):
    #     """Test class to get only quadrilateral faces."""
    #     in_faces = self.mesh.in_faces
    #     tri_faces, _ = self.mesh.screen_faces_by_verts(in_faces)
    #     self.assertEqual(len(tri_faces),  11)
    #
    # def test_screening_faces_gets_only_quadrilateral_boundary_faces(self):
    #     """Test class to get only quadrilateral faces."""
    #     in_faces = self.mesh.b_faces
    #     _, in_quad_faces = self.mesh.screen_faces_by_verts(in_faces)
    #     self.assertEqual(len(in_quad_faces),  6)
    #
    # def test_screening_faces_gets_only_triangular_boundary_faces(self):
    #     """Test class to get only quadrilateral faces."""
    #     in_faces = self.mesh.b_faces
    #     tri_faces, _ = self.mesh.screen_faces_by_verts(in_faces)
    #     self.assertEqual(len(tri_faces),  10)
    #
    # def test_if_permeability_tensor_is_assigned(self):
    #     """Test if permeability tensor is being assigned for all volumes."""
    #     all_volumes = self.mesh.all_volumes
    #     K_1 = [1.0, 0.0, 0.0,
    #            0.0, 1.0, 0.0,
    #            0.0, 0.0, 1.0]
    #     for volume in all_volumes:
    #         perm = self.mesh.M.permeability[[volume]][0]
    #         self.assertListEqual(list(perm), K_1)
    #
    # def test_get_dirichlet_faces(self):
    #     """Test if Dirichlet BC is implemented."""
    #     d_faces = self.mesh.M.faces.flag[101]
    #     dirichlet_faces_pressure = self.mesh.M.dirichlet_faces
    #     for face in d_faces:
    #         b_pressure = dirichlet_faces_pressure[[face]]
    #         self.assertEqual(b_pressure, 0.)
    #
    # def test_position_left_and_right_volumes_for_triangular_intern_faces(self):
    #     """Test if all volumes are properly oriented with the normal."""
    #     in_faces = self.mesh.in_faces
    #     tri_faces = self.mesh.screen_faces_by_verts(in_faces)[0]
    #     (left_volumes,
    #      right_volumes) = self.mesh.get_left_and_right_volumes(tri_faces)
    #     N_IJK = self.mesh.construct_face_vectors(tri_faces)[0]
    #     vector_left_to_right = (self.mesh.M.volumes.center[left_volumes]
    #                             - self.mesh.M.volumes.center[right_volumes])
    #     is_positive = np.sum(vector_left_to_right * N_IJK, axis=1)
    #     self.assertFalse(list(np.flatnonzero(is_positive < 0)))
    #
    # def test_position_left_and_right_volumes_for_quad_intern_faces(self):
    #     """Test if all volumes are properly oriented with the normal."""
    #     in_faces = self.mesh.in_faces
    #     quad_faces = self.mesh.screen_faces_by_verts(in_faces)[1]
    #     (left_volumes,
    #      right_volumes) = self.mesh.get_left_and_right_volumes(quad_faces)
    #     N_IJK = self.mesh.construct_face_vectors(quad_faces)[0]
    #     vector_left_to_right = (self.mesh.M.volumes.center[left_volumes]
    #                             - self.mesh.M.volumes.center[right_volumes])
    #     is_positive = np.sum(vector_left_to_right * N_IJK, axis=1)
    #     self.assertFalse(list(np.flatnonzero(is_positive < 0)))
    #
    # def test_get_normal_face_area_vector_points_outward(self):
    #     """Test if bondary volumes have their normal outward oriented."""
    #     quad_faces = self.mesh.screen_faces_by_verts(self.mesh.b_faces)[1]
    #     N_IJK = self.mesh.construct_face_vectors(quad_faces, boundary=True)[0]
    #     adj_vol = self.mesh.M.faces.bridge_adjacencies(quad_faces, 2, 3)
    #     i = self.mesh.get_position_IJK_verts(quad_faces)[0]
    #     vector_left_to_right = (self.mesh.M.nodes.coords[i]
    #                             - self.mesh.M.volumes.center[adj_vol])
    #     is_positive = np.sum(vector_left_to_right * N_IJK, axis=1)
    #     self.assertFalse(list(np.flatnonzero(is_positive < 0)))
    #
    # def test_calculate_area_vector(self):
    #     b_faces = self.mesh.b_faces
    #     tri_faces, quad_faces = self.mesh.screen_faces_by_verts(b_faces)
    #     face_area_quad = self.mesh.get_area(quad_faces)
    #     face_area_tri = self.mesh.get_area(tri_faces)
    #     self.assertEqual(np.sum(face_area_tri) + np.sum(face_area_quad), 6.)
    #
    # def test_calculate_additional_geometric_information(self):
    #     in_faces = self.mesh.in_faces
    #     quad_faces = self.mesh.screen_faces_by_verts(in_faces)[1]
    #     LR, h_L, h_R = self.mesh.get_additional_vectors_and_height(quad_faces)
    #     self.assertEqual(LR[0][0], h_L + h_R)
    #
    # @unittest.skip("Test not ready. Still missing a test case assertion")
    # def test_calculate_additional_geometric_information_boundary(self):
    #     b_faces = self.mesh.b_faces
    #     quad_faces = self.mesh.screen_faces_by_verts(b_faces)[1]
    #     LJ, h_L = self.mesh.get_additional_vectors_and_height(quad_faces,
    #                                                           boundary=True)
    #     pass
    #
    # def test_calculate_volumes(self):
    #     all_vol_volumes = self.mesh.get_volume(self.mesh.all_volumes)
    #     # print(all_vol_volumes)
    #     self.assertEqual(sum(all_vol_volumes), 1.)
    #
    # def test_run_preprocessor(self):
    #     try:
    #         self.mesh.run_preprocessor()
    #     except:
    #         self.fail("mesh.preprocessor did not run.")


if __name__ == "__main__":
    unittest.main()



#__all__ = ['meshManager']
#import meshManager

#from meshUtil import meshManager


#shManager.py


print('configTools intialized')
#from meshManager.py import MeshManager as ms

#from .Structured import Preprocessor as Structured
#from .Structured2D import Preprocessor as Structured2D



# Procedure to get the relevant vertices for the geometric module in each type of elements

from pymoab import core, topo_util, rng


mb = core.Core()
mesh_topo_util = topo_util.MeshTopoUtil(mb)

# Piramide


def get_pyramid_adjacencies(vertices, edges):

    # Creating used ranges
    aux1 = rng.Range()
    aux2 = rng.Range()
    base_node = rng.Range()
    top_node = rng.Range()

    # Discovering the top node of the pyramid
    for e in vertex_handles:
        adj = mesh_topo_util.get_bridge_adjacencies(e, 1, 0)
        if len(adj) == 4:
            top_node.insert(e)

    # Defines the base nodes
    base_nodes = rng.subtract(vertex_handles, top_node)

    # Getting the base adjacencies of the point of reference
    aux = mesh_topo_util.get_bridge_adjacencies(base_nodes[0], 1, 0)
    aux = rng.subtract(top_node, aux)

    aux1.insert(aux[0])
    aux2.insert(aux[1])
    base_node.insert(base_nodes[0])
    coord1 = mbcore.get_coords(base_node)
    coord2 = mbcore.get_coords(aux1)
    coord3 = mbcore.get_coords(aux2)
    coord4 = mbcore.get_coords(top_node)
    order = ([coord1, coord2, coord3, coord4])

    return order

def get_hexahedro_adjacencies(vertices, edges):

    # Creating used ranges
    aux1 = rng.Range()
    aux2 = rng.Range()
    base_nodes = rng.Range()
    top_node = rng.Range()

    # Discovering the top node of the pyramid
    for e in vertex_handles:
        adj = mesh_topo_util.get_bridge_adjacencies(e, 1, 0)
        if len(adj) == 4:
            top_node.insert(e)

    # Defines the base nodes
    base_nodes = rng.subtract(vertex_handles, top_node)

    # Getting the base adjacencies of the point of reference
    aux = mesh_topo_util.get_bridge_adjacencies(base_nodes[0], 1, 0)
    aux = rng.subtract(top_node, aux)

    aux1.insert(aux[0])
    aux2.insert(aux[1])
    base_node.insert(base_nodes[0])
    coord1 = mbcore.get_coords(base_node)
    coord2 = mbcore.get_coords(aux1)
    coord3 = mbcore.get_coords(aux2)
    coord4 = mbcore.get_coords(top_node)
    order = ([coord1, coord2, coord3, coord4])

    return order





import numpy as np
import yaml
import os
from ..meshHandle.finescaleMesh import FineScaleMesh
from ..meshHandle.configTools.configClass import variableInit
from pymoab import types
import copy


class partitionManager(object):
    def __init__(self, M, config_object):
        scheme = config_object.tree['Scheme']
        if scheme == 'smart':
            self.func = self.smart
            if config_object.tree['Smart']['path'] == 'default':
                file_path = os.getcwd() + '/mesh/coarse/' + \
                    config_object.tree['Smart']['file']
            else:
                file_path = config_object.tree['Smart']['path'] + '/' + \
                    config_object.tree['Smart']['file']
            self.arg = [file_path]
            self.partitioner = smartPartition(M)
        elif scheme == 'simple':
            self.func = self.simple
            self.partitioner = simplePartition(M)
            self.arg = config_object.tree['Simple']['nx'], config_object.tree['Simple']['ny'], config_object.tree['Simple']['nz']
        elif scheme == 'parallel':
            print('Looking for Parallel Partition')
            self.func = self.parallel()
        else:
            print('Scheme ' + scheme + ' not defined.')

    def __call__(self):
        return self.run()

    def run(self, *args):
        if len(args) == 0:
            return self.func(*self.arg)
        else:
            return self.func(*args)

    def parallel(self):
        return ['parallel', None]

    def smart(self, file_path):
        return self.partitioner(file_path)

    def simple(self, nx, ny, nz):
        return self.partitioner(nx, ny, nz)


class smartPartition(object):
    def __init__(self,M):
        self.M = M

    def __call__(self, file_path):
        return self.run(file_path)

    def run(self, file_path):
        print('Creating Coarse Scale Forming Primal Grid')
        self.variable_entries = variableInit(empty=True)
        self.variable_entries.add_entry('pressure', 'volumes', 1, 'float')
        self.primal = FineScaleMesh(file_path, var_config=copy.deepcopy(self.variable_entries))
        #import pdb; pdb.set_trace()
        self.dual = self.create_dual()

    def create_dual(self):
        nodes_coords = self.primal.nodes.center[:]
        edges_center = self.primal.edges.center[:]
        faces_center = self.primal.faces.center[:]
        volumes_center = self.primal.volumes.center[:]
        sizes = np.array([len(nodes_coords), len(edges_center), len(faces_center), len(volumes_center)])
        sizes = np.cumsum(sizes)
        sizes = np.concatenate((np.array([0]), sizes))
        tag = lambda ind, type: (sizes[type] + ind +1)
        all_tetra = np.array([[], [], [], []]).T
        for vol in self.primal.volumes.all:
            faces = self.primal.volumes.adjacencies[vol]
            edges = self.primal.volumes._adjacencies(vol, dim_tag=1)
            nodes = self.primal.volumes.connectivities[vol]
            for edge in edges.T:
                adj_faces = np.intersect1d(self.primal.edges.bridge_adjacencies(edge, interface="edges",target="faces"), faces)
                node_edge = self.primal.edges.connectivities[edge.T].ravel()
                tetras = np.array([[tag(edge, 1), tag(node_edge[0], 0), tag(adj_faces[0], 2), tag(vol, 3)],
                                   [tag(edge, 1), tag(node_edge[0], 0), tag(adj_faces[1], 2), tag(vol, 3)],
                                   [tag(edge, 1), tag(node_edge[1], 0), tag(adj_faces[0], 2), tag(vol, 3)],
                                   [tag(edge, 1), tag(node_edge[1], 0), tag(adj_faces[1], 2), tag(vol, 3)]])
                all_tetra = np.vstack((all_tetra, tetras))
        print('Creating Coarse Scale Forming Dual Grid')
        dual = FineScaleMesh(mesh_file=None, dim=3, var_config=self.variable_entries)
        dual.core.mb.create_vertices(np.vstack((nodes_coords, edges_center, faces_center, volumes_center)))
        for tetra in all_tetra:
            dual.core.mb.create_element(types.MBTET, tetra.ravel().astype("uint64"))
        dual.core.run()
        dual.run()
        return dual


class simplePartition(object):
    def __init__(self, M):
        # num_of_vol, rx,ry,rz ,nx = 3, ny = 3, nz =3
        self.M = M

    def __call__(self, nx=4, ny=4, nz=4):
        return self.scheme(nx, ny, nz)

    def scheme(self, nx=4, ny=4, nz=4):
        centerCoord = self.M.volumes.center[:]
        num_of_vol = len(self.M)
        rx, ry, rz = self.M.rx, self.M.ry, self.M.rz
        if (rz[1] == 0) & (rz[0] == 0):
            nz = 1
            rz = (-1,1)
        box = np.array([0, (rx[1] - rx[0])/nx, 0,
              (ry[1] - ry[0]) /ny, 0,(rz[1]- rz[0])/(nz+0)]).reshape(3,2)
        cent_coord_El1 = box.sum(axis =1)/2
        tag = np.zeros(num_of_vol).astype("int")
        coarseCenters = np.zeros((nx*ny*nz,3))
        index = 0
        init_coords = np.array([rx[0],ry[0],rz[0]])
        for x in range(nx):
            for y in range(ny):
                for z in range(nz):
                    inc = np.multiply(box[:,1], np.array([x,y,z]))

                    #cent = cent_coord_El1 + inc
                    coarseCenters[index] = cent_coord_El1 + inc
                    # pdb.set_trace()

                    #inc = np.array([(nx) * x, (ny) * y, (nz) * z])
                    boxMin = box[:,0] + inc + init_coords
                    boxMax = box[:,1] + inc + init_coords
                    point = self.check_in_box(centerCoord,x=(boxMin[0], boxMax[0]), y=(boxMin[1], boxMax[1]) , z=(boxMin[2], boxMax[2]))
                    tag[point] = index
                    index += 1
        return self.tag_adjust(tag, coarseCenters)

    def check_in_box(self,coords, x , y, z):
        tag1 = (coords[:,0] > x[0]) & (coords[:,0] < x[1])
        tag2 = (coords[:,1] > y[0]) & (coords[:,1] < y[1])
        tag3 = (coords[:,2] > z[0]) & (coords[:,2] < z[1])
        return tag1 & tag2 & tag3

    def tag_adjust(self, tag, coarseCenter):
        fineTag =  tag
        elementsOriginal = [*set(tag)]
        elementsNovo = [*set(range(len(elementsOriginal)))]
        elementsMissing = set(range(len(coarseCenter))) - set(elementsOriginal)
        for elo, eln in zip(elementsOriginal,elementsNovo):
            if elo != eln:
                pointer = (tag == elo)
                fineTag[pointer] = eln
        return fineTag.astype(int) , np.delete(coarseCenter, [*elementsMissing], axis = 0)



"""
Use of Pymoab methods to read and manage the input mesh
"""
from pymoab import core, types, rng, topo_util
from pymoab import skinner as sk
import numpy as np
import yaml
import time



class CoreMoab:
    def __init__(self, mesh_file=None, dim=3):
        self.dimension = dim
        self.mb = core.Core()
        self.mtu = topo_util.MeshTopoUtil(self.mb)
        if mesh_file is not None:
            self.mb.load_file(mesh_file)
            self.run()

    def run(self):
        self.root_set = self.mb.get_root_set()
        self.father_root_set = self.root_set
        self.level = 0
        self.all_volumes = self.mb.get_entities_by_dimension(0, 3)
        self.all_nodes = self.mb.get_entities_by_dimension(0, 0)
        self.mtu.construct_aentities(self.all_nodes)
        #self.mtu.construct_aentities(self.all_volumes)
        self.all_faces = self.mb.get_entities_by_dimension(0, 2)
        self.all_edges = self.mb.get_entities_by_dimension(0, 1)
        self.handleDic = {}
        [self.boundary_nodes, self.boundary_edges, self.boundary_faces, self.boundary_volumes] = self.skinner_operation()
        self.internal_nodes = rng.subtract(self.all_nodes, self.boundary_nodes)
        self.internal_edges = rng.subtract(self.all_edges, self.boundary_edges)
        self.internal_faces = rng.subtract(self.all_faces, self.boundary_faces)
        self.internal_volumes = rng.subtract(self.all_volumes, self.boundary_volumes)
        self.id_name = "GLOBAL_ID"
        self.father_id_name = self.id_name
        self.init_id()
        #self.check_integrity()
        # self.create_id_visualization()
        # self.flag_dic = {}
        [self.flag_list, self.flag_dic] = self.read_flags()

        # self.create_flag_visualization()

        # swtich on/off
        self.parallel_meshset = self.create_parallel_meshset()
        self.create_parallel_tag()

    def init_id(self):
        self.create_tag_handle(self.id_name, data_size = 1, data_text = "int", data_density = "sparse")
        # create volume ids
        self.set_data(self.id_name, np.arange(len(self.all_volumes)))
        # create face ids
        self.set_data(self.id_name, np.arange(len(self.all_faces)), range_el=self.all_faces)
        # create edges ids
        self.set_data(self.id_name, np.arange(len(self.all_edges)), range_el=self.all_edges)
        # create nodes ids
        self.set_data(self.id_name, np.arange(len(self.all_nodes)), range_el=self.all_nodes)

    def skinner_operation(self):
        self.skin = sk.Skinner(self.mb)
        print("Entering skinner test")

        if self.dimension == 3:
            faces_on_skin_handles = self.skin.find_skin(self.root_set, self.all_volumes)
            edges_on_skin_handles = self.mtu.get_bridge_adjacencies(faces_on_skin_handles, 2, 1)
            nodes_on_skin_handles = self.mtu.get_bridge_adjacencies(faces_on_skin_handles, 2, 0)
            volumes_on_skin_handles = self.mtu.get_bridge_adjacencies(faces_on_skin_handles, 0, 3)
        elif self.dimension == 2:
            edges_on_skin_handles = self.skin.find_skin(self.root_set, self.all_faces)
            nodes_on_skin_handles = self.mtu.get_bridge_adjacencies(edges_on_skin_handles, 1, 0)
            faces_on_skin_handles = self.mtu.get_bridge_adjacencies(edges_on_skin_handles, 0, 2)
            volumes_on_skin_handles = self.mtu.get_bridge_adjacencies(edges_on_skin_handles, 0, 3) #empty

        print("Skinning Operation Successful")
        return [nodes_on_skin_handles, edges_on_skin_handles, faces_on_skin_handles, volumes_on_skin_handles]


    def check_integrity(self):
        # check if the mesh contains
        check_list = [len(self.all_nodes), len(self.all_edges), len(self.all_faces), len(self.all_volumes)]
        list_words = ['Nodes', "Edges", "Faces", "Volumes"]
        print("Checking mesh integrity:")
        index = 0
        if self.dimension == 2:
            list_words = list_words[:-1]
            check_list = check_list[:-1]
        for entity in check_list:
            if entity > 0:
                print(list_words[index] + " successfully imported")
            else:
                print("------------------------------\nError creating \n" +
                      list_words[index] + " was not imported")
            index += 1

    def create_parallel_meshset(self):
        partition_volumes = []
        parallel_tag = []
        try:
            parallel_tag = self.mb.tag_get_handle("PARALLEL_PARTITION")
            flag = False
        except:
            print("Parallel Partition Tag not found \nAborting creating parallel partition entities.")
            flag = True
        if not flag:
            print("Parallel Partition Tag detected \nCreating parallel partitions entities")
            self.handleDic["PARALLEL_PARTITION"] = parallel_tag
            parallel_sets = self.mb.get_entities_by_type_and_tag(
                0, types.MBENTITYSET, np.array(
                (parallel_tag,)), np.array((None,)))
            self.create_tag_handle("Parallel", data_size=1, data_text="int")
            # partition_volumes = []
            for set_el in parallel_sets:
                num_tag = self.read_data("PARALLEL_PARTITION", range_el=set_el)[0, 0]
                # self.set_data("Parallel", np.ones(len(set_el))*num_tag,  range_el=set_el)
                list_entity = [self.mb.get_entities_by_dimension(set_el, 0), self.mb.get_entities_by_dimension(set_el, 1),
                               self.mb.get_entities_by_dimension(set_el, 2), self.mb.get_entities_by_dimension(set_el, 3)]
                # print([num_tag, list_entity])
                partition_volumes.append(list_entity)
        return partition_volumes

    def create_parallel_tag(self):
        k = 0
        for sets in self.parallel_meshset:
            for dim in sets:
                if len(dim) != 0:
                    self.set_data("Parallel", k * np.ones(len(dim)).astype(int), range_el=dim)
            k += 1

    def read_flags(self):
        physical_tag = self.mb.tag_get_handle("MATERIAL_SET")
        physical_sets = self.mb.get_entities_by_type_and_tag(
            0, types.MBENTITYSET, np.array(
            (physical_tag,)), np.array((None,)))
        self.handleDic["MATERIAL_SET"] = physical_tag
        flag_list = np.array([])
        flag_dic = {}
        for set in physical_sets:
            bc_flag = self.read_data("MATERIAL_SET", range_el=set)[0, 0]
            flag_list = np.append(flag_list, bc_flag)
            list_entity = [self.mb.get_entities_by_dimension(set, 0), self.mb.get_entities_by_dimension(set, 1),
                           self.mb.get_entities_by_dimension(set, 2), self.mb.get_entities_by_dimension(set, 3)]
            flag_dic[bc_flag] = list_entity
        return np.sort(flag_list), flag_dic

    # def access_meshset(self, handle):
    #     # returns the entities contained inside a give meshset handle
    #     # ie: for a meshset handle the entities inside are returned
    #     temp_range = []
    #     for el in range(self.dimension+1):
    #         sub_el = (self.mb.get_entities_by_dimension(handle, el))
    #         temp_range.append(sub_el)
    #     temp_range.append(self.mb.get_entities_by_dimension(handle, 11))
    #     return temp_range

    # def access_handle(self,handle):
    #     # input: range of handles of different dimensions
    #     # returns all entities with d-1 dimension the comprises the given range
    #     # ie: for a volume, the faces, for a face the edges and for an edge the points.
    #     #
    #     vecdim = self.check_range_by_dimm(handle)
    #     # pdb.set_trace()
    #     all_adj = np.array([np.array(self.mb.get_adjacencies(el_handle, dim-1)) for dim, el_handle in zip(vecdim,handle)])
    #     #unique_adj = np.unique(np.ma.concatenate(all_adj)).astype("uint64")
    #     # print(handle,all_adj)
    #     unique_adj = np.unique(np.concatenate(all_adj)).astype("uint64")
    #     return rng.Range(unique_adj)

    def create_tag_handle(self, name_tag, data_size, data_text="float", data_density="dense"):
        if data_density == "dense":
            data_density = types.MB_TAG_DENSE
        elif data_density == "sparse":
            data_density = types.MB_TAG_SPARSE
        elif data_density == "bit":
            data_density = types.MB_TAG_BIT
        else:
            print("Please define a valid tag type")
        if data_text == 'float':
            data_type = types.MB_TYPE_DOUBLE
        elif data_text == "int":
            data_type = types.MB_TYPE_INTEGER
        elif data_text == "bool":
            data_type = types.MB_TYPE_BIT
        try:
            handle = self.handleDic[name_tag]
        except KeyError:
            handle = self.mb.tag_get_handle(name_tag, data_size, data_type, data_density, True)
            self.handleDic[name_tag] = handle

    def read_data(self, name_tag, index_vec = np.array([]), range_el = None):
        if range_el is None:
            range_el = self.all_volumes
        if index_vec.size > 0:
            range_el = range_el[index_vec]
        try:
            handle_tag = self.handleDic[name_tag]
            return self.mb.tag_get_data(handle_tag, range_el)
        except KeyError:
            print("Tag not found")

    # def range_index(self, vec_index, range_handle=None):
    #     if range_handle is None:
    #         range_handle = self.all_volumes
    #     if vec_index.dtype == "bool":
    #         vec = np.where(vec_index)[0]
    #     else:
    #         vec = vec_index.astype("uint")
    #     handles = np.asarray(range_handle)[vec.astype("uint")].astype("uint")
    #     return rng.Range(handles)

    def set_data(self, name_tag, data, index_vec=np.array([]), range_el=None):
        if range_el is None:
            range_el = self.all_volumes
        if index_vec.size > 0:
            range_el = range_el[index_vec]
        handle_tag = self.handleDic[name_tag]
        self.mb.tag_set_data(handle_tag, range_el, data)

    def check_range_by_dimm(self, handle):
        # INPUT: handle or range
        # OUTPUT: a vector with the same size as the input handle with the following classification
        # 0 - nodes , 1 -edges, 2 - faces 3 - volumes, 4- meshset
        handle_int = np.asarray(handle).astype("uint64")
        type_list = np.array([self.mb.type_from_handle(el) for el in handle_int])
        handle_classification = np.zeros(len(handle))
        nodetype = type_list == types.MBVERTEX
        edgetype = type_list == types.MBEDGE
        facetype = (type_list == types.MBTRI) | (type_list == types.MBQUAD) | (type_list == types.MBPOLYGON)
        volumetype = (type_list == types.MBTET) | (type_list == types.MBPYRAMID) | (type_list == types.MBPRISM) | \
                      (type_list == types.MBKNIFE) | (type_list == types.MBHEX) | (type_list == types.MBPOLYHEDRON)
        meshsettype = type_list == types.MBENTITYSET
        handle_classification[nodetype] = 0
        handle_classification[edgetype] = 1
        handle_classification[facetype] = 2
        handle_classification[volumetype] = 3
        handle_classification[meshsettype] = 11
        return handle_classification.astype("uint64")

    def filter_range(self, handle, filter_vec):
        # INPUT: handle or range
        # OUTPUT: handles in the range of the dimension in args
        # 0 - nodes , 1 -edges, 2 - faces 3 - volumes, 4- meshset
        handle_int = np.asarray(handle).astype("uint64")
        if filter_vec.dtype == bool:
            return rng.Range(handle_int[filter_vec])
        else:
            return rng.Range(handle_int[filter_vec.astype("uint64")])

    def filter_handle_by_dimension(self, handle, *args):
        # INPUT: handle or range
        # OUTPUT: handles in the range of the dimension in args
        # 0 - nodes , 1 -edges, 2 - faces 3 - volumes, 4- meshset
        handle_int = np.asarray(handle).astype("uint64")
        vec_classification = self.check_range_by_dimm(handle)
        test_elem = np.array([*args])
        return rng.Range(handle_int[np.isin(vec_classification,test_elem)])

    @staticmethod
    def range_merge(*args):
        range_merged = rng.Range()
        for arg in args:
                range_merged.merge(arg)
        return range_merged

    def print(self, file=None, extension=".h5m", case = None,  config_input="input_cards/print_settings.yml"):
        if case is None:
            case = ''
        text =  file
        folder = "results/" + case
        with open(config_input, 'r') as f:
            data = yaml.safe_load(f)
        nodes = data['nodes']
        edges = data['edges']
        faces = data['faces']
        volumes = data['volumes']
        all_entities = data['all entities']
        m1 = self.mb.create_meshset()
        self.mb.add_entities(m1, self.all_nodes)
        m2 = self.mb.create_meshset()
        self.mb.add_entities(m2, self.all_faces)
        self.mb.remove_entities(m2, self.all_nodes)
        m3 = self.mb.create_meshset()
        self.mb.add_entities(m3, self.all_volumes)
        m4 = self.mb.create_meshset()
        self.mb.add_entities(m4, self.all_edges)
        if text is None:
            text = "output"
        text1 = text + "-nodes" + extension
        text2 = text + "-face" + extension
        text3 = text + "-volume" + extension
        text4 = text + "-edges" + extension
        text5 = text + "-all" + extension
        if folder is not None:
            import os
            if not os.path.exists(folder):
                os.mkdir(folder)
            text = folder + '/' + text
            text1 = folder + '/' + text1
            text2 = folder + '/' + text2
            text3 = folder + '/' + text3
            text4 = folder + '/' + text4
            text5 = folder + '/' + text5
        if nodes != 0:
            self.mb.write_file(text1, [m1])
        if faces != 0:
            self.mb.write_file(text2, [m2])
        if self.dimension == 3 and volumes != 0:
            self.mb.write_file(text3, [m3])
        if edges != 0:
            self.mb.write_file(text4, [m4])
        if all_entities != 0:
            self.mb.write_file(text5)



"""
Generator of mesh entities and tags
"""

import pdb
import numpy as np
from pymoab import types, rng, topo_util
from ..geoUtil import geoTools as gtool
#from preprocessor.meshHandle.configTools.configClasses import variableInit


class GetItem(object):
    def __init__(self, adj):
        self.fun = adj

    def __call__(self, item):
        return self.fun(item)

    def __getitem__(self, item):
        return self.fun(item)


class MeshEntities(object):
    def __init__(self, core, entity_type):
        self.mb = core.mb
        self.mtu = core.mtu
        self.meshset = core.root_set
        self.nodes = core.all_nodes
        self.num = {"nodes": 0, "node": 0, "edges": 1, "edge": 1, "faces": 2, "face": 2, "volumes": 3, "volume": 3,
                    0: 0, 1: 1, 2: 2, 3: 3}
        string = {0: "nodes", 1: "edges", 2: "faces", 3: "volumes"}
        entity_num = self.num[entity_type]
        list_type = [(core.all_nodes, core.internal_nodes, core.boundary_nodes),
                        (core.all_edges, core.internal_edges, core.boundary_edges),
                        (core.all_faces, core.internal_faces, core.boundary_faces),
                        (core.all_volumes, core.internal_volumes, core.boundary_volumes)]
        self.list_all = [list_type[0][0], list_type[1][0], list_type[2][0], list_type[3][0]]
        self.level = core.level
        if self.level == 0:
            self.id_name = "GLOBAL_ID"
            self.father_id_name = "GLOBAL_ID"
            self.id_name = "GLOBAL_ID"
        elif self.level == 1:
            self.father_id_name = core.father_core.id_name
            self.id_name = "LOCAL_ID_L" + str(core.level) + "-" + str(core.coarse_num)
        else:
            self.father_id_name = core.father_core.id_name
            self.id_name = self.father_id_name + str("L") + str(self.level) + "-" + str(self.coarse_num)

        (self.elements_handle, self.internal_elements, self.boundary_elements), self.vID = list_type[entity_num], entity_num
        # self.internal_elements_array = self.internal_elements.get_array()
        self.entity_type = string[entity_num]
        self.tag_handle = core.handleDic[self.id_name]
        self.global_handle = core.handleDic['GLOBAL_ID']
        self.father_handle = core.handleDic[self.father_id_name]
        if self.vID == 0:
            self.adjacencies = GetItem(self._adjacencies_for_nodes)
            self.coords =  GetItem(self._coords)
        else:
            self.adjacencies = GetItem(self._adjacencies)
            self.connectivities = GetItem(self._connectivities)
        self.classify_element = GetItem(self._classify_element)
        self.center = GetItem(self._center)
        # self.global_id = GetItem(self._global_id)
        # self.father_id = GetItem(self._father_id)
        if (self.vID == 1) & (core.dimension == 2):
            self.normal = GetItem(self._normal)
        elif (self.vID == 2) & (core.dimension == 3):
            self.normal = GetItem(self._normal)
        # initialize specific flag dic in accordance with type of the object create
        self.flag = {key: self.read(value[self.vID]) for key, value in core.flag_dic.items()
                     if value[self.vID].empty() is not True}
        # print("Mesh Entity type {0} successfully initialized".format(entity_type))

    def bridge_adjacencies(self, index, interface, target):
        if isinstance(index, np.int64) or isinstance(index, int):
            el_handle = np.array([self.elements_handle[index]])
        elif not isinstance(index, np.ndarray) and index is not None:
            el_handle = self.elements_handle[index].get_array()
        else:
            el_handle = self.elements_handle.get_array(index)
        if self.level==0:
            return self.mtu.get_ord_bridge_adjacencies(el_handle, self.num[interface], self.num[target], self.mb, self.tag_handle)
        return self.mtu.get_ord_bridge_adjacencies(el_handle, self.num[interface], self.num[target], self.mb, self.tag_handle, self.list_all[self.num[target]], self.level)

    def _coords(self, index):
        if isinstance(index, np.int64) or isinstance(index, int):
            el_handle = np.array([self.nodes[index]])
        elif not isinstance(index, np.ndarray) and index is not None:
            el_handle = self.nodes[index]
        else:
            el_handle = self.nodes.get_array(index)
        if len(el_handle) == 1:
            return self.mb.get_coords(el_handle)
        return np.reshape(self.mb.get_coords(el_handle),(-1,3))

    # def _global_id(self, index):
    #     range_vec = self.create_range_vec(index)
    #     elements_handle = self.range_index(range_vec)
    #     return self.mb.tag_get_data(self.global_handle, elements_handle).ravel()
    # def _father_id(self, index):
    #     range_vec = self.create_range_vec(index)
    #     elements_handle = self.range_index(range_vec)
    #     return self.mb.tag_get_data(self.father_handle, elements_handle).ravel()

    def _adjacencies_for_nodes(self, index):
        return index

    def _adjacencies(self, index, flag_nodes=False, dim_tag=None):
        if dim_tag is None:
            if not flag_nodes:
                dim_tag = self.vID - 1
            else:
                dim_tag = 0
        if isinstance(index, np.int64) or isinstance(index, int):
            el_handle = np.array([self.elements_handle[index]])
        elif not isinstance(index, np.ndarray) and index is not None:
            el_handle = self.elements_handle[index].get_array()
        else:
            el_handle = self.elements_handle.get_array(index)
        return self.mb.get_ord_adjacencies(el_handle, dim_tag, tag_handle = self.tag_handle)

    def _center(self,index):
        if self.vID == 0:
            centers = self._coords(index)
            return centers
        elif self.vID == 1:
            edges_adj = self.connectivities[index]
            if edges_adj.ndim == 1:
                return 0.5*(self._coords(edges_adj[0]) + self._coords(edges_adj[1]))
            v0 = np.array([edges_adj[i][0] for i in range (edges_adj.shape[0])])
            v1 = np.array([edges_adj[i][1] for i in range (edges_adj.shape[0])])
            centers = 0.5 * (self._coords(v0) + self._coords(v1))
            return centers
        elif self.vID == 2 or self.vID == 3:
            if isinstance(index, np.int64) or isinstance(index, int):
                el_handle = np.array([self.elements_handle[index]])
            elif not isinstance(index, np.ndarray) and index is not None:
                el_handle = self.elements_handle[index].get_array()
            else:
                el_handle = self.elements_handle.get_array(index)
            adj = self.mb.get_ord_connectivity(el_handle, tag_opt = False)
            if adj.ndim==1:
                return gtool.get_average(np.reshape(self.mb.get_coords(adj),(-1,3)))
            centers = np.empty((adj.shape[0],3))
            for i in range(adj.shape[0]):
                centers[i] = gtool.get_average(np.reshape(self.mb.get_coords(adj[i]),(-1,3)))
                #pdb.set_trace()
            return centers
        return None

    def _normal(self,index):

        #normal_vec = np.zeros(( np.shape(range_vec)[0],3 ))
        adj = self.connectivities[index]
        if adj.ndim==1:
            if self.vID == 1:
                return gtool.normal_vec_2d(self._coords(adj[0]).reshape(1,3), self._coords(adj[1]).reshape(1,3))[0]
            elif self.vID == 2:
                return gtool.normal_vec(self._coords(adj[0]).reshape(1,3),self._coords(adj[1]).reshape(1,3),self._coords(adj[2]).reshape(1,3))[0]
            return
        v0 = np.array([adj[i][0] for i in range (adj.shape[0])])
        v1 = np.array([adj[i][1] for i in range (adj.shape[0])])
        if self.vID == 1:
            return gtool.normal_vec_2d(self._coords(v0),self._coords(v1))

            #edges_adj = self.connectivities[range_vec]
            #centers  = 0.5* (self._coords(edges_adj[:,0]) + self._coords(edges_adj[:,1]))
        elif self.vID == 2:
            v2 = np.array([adj[i][2] for i in range (adj.shape[0])])
            return  gtool.normal_vec(self._coords(v0),self._coords(v1),self._coords(v2))

    def _connectivities(self,index):
        if isinstance(index, np.int64) or isinstance(index, int):
            el_handle = np.array([self.elements_handle[index]])
        elif not isinstance(index, np.ndarray) and index is not None:
            el_handle = self.elements_handle[index].get_array()
        else:
            el_handle = self.elements_handle.get_array(index)
        return self.mb.get_ord_connectivity(el_handle, tag_handle = self.tag_handle)

    def create_range_vec(self, index):
        range_vec = None
        if isinstance(index, int) or isinstance(index, np.integer):
            range_vec = np.array([index]).astype("uint")
        elif isinstance(index, np.ndarray):
            if index.dtype == "bool":
                range_vec = np.where(index)[0]
            else:
                range_vec = index
        elif isinstance(index, slice):
            start = index.start
            stop = index.stop
            step = index.step
            if start is None:
                start = 0
            if stop is None:
                stop = len(self)
            if step is None:
                step = 1
            if start < 0:
                start = len(self) + start + 1
            if stop < 0:
                stop = len(self) + stop + 1
            range_vec = np.arange(start, stop, step).astype('uint')
        elif isinstance(index, list):
            range_vec = np.array(index)
        return range_vec

    def init_partition(self):
        config = self.read_config('msCoarse.yml')
        particionador_type = config["Partitioner Scheme"]
        specific_attributes = config["Coarsening"]
        if particionador_type != '0':
            if self.dim == 3:
                partition = MoabVariable(self.core,data_size=1,var_type= "volumes",  data_format="int", name_tag="Partition",
                                             data_density="sparse")
                name_function = "scheme" + particionador_type
                used_attributes = []
                used_attributes.append(specific_attributes[0]["nx"])
                used_attributes.append(specific_attributes[1]["ny"])
                used_attributes.append(specific_attributes[2]["nz"])
                [partition[:],coarse_center]  = getattr(algoritmo, name_function)(self.volumes.center[:],
                           len(self), self.rx, self.ry, self.rz,*used_attributes)
            elif self.dim == 2:
                partition = MoabVariable(self.core,data_size=1,var_type= "faces",  data_format="int", name_tag="Partition",
                                             data_density="sparse")
                name_function = "scheme" + particionador_type
                specific_attributes = config["Coarsening"]
                used_attributes = []
                used_attributes.append(specific_attributes[0]["nx"])
                used_attributes.append(specific_attributes[1]["ny"])
                [partition[:],coarse_center]  = getattr(algoritmo, name_function)(self.faces.center[:],
                           len(self), self.rx, self.ry, self.rz,*used_attributes)
            return partition

    def _classify_element(self, index):
        range_vec = self.create_range_vec(index)
        range = self.range_index(range_vec)
        type_list = np.array([self.mb.type_from_handle(el) for el in range])
        return type_list

    def range_index(self, vec_index, flag_nodes=False):
        if not flag_nodes:
            range_handle = self.elements_handle
        else:
            range_handle = self.nodes
        if vec_index.dtype == "bool":
            vec = np.where(vec_index)[0]
        else:
            vec = vec_index.astype("uint")
        handles = np.asarray(range_handle)[vec.astype("uint")].astype("uint")
        return handles
        # return rng.Range(handles)

    def __str__(self):
        string = "{0} object \n Total of {1} {0} \n {2}  boundary {0} \n {3} internal {0}".format(self.entity_type,
            len(self.elements_handle), len(self.boundary_elements), len(self.internal_elements))
        return string

    def __len__(self):
        return len(self.elements_handle)

    def __call__(self):
        return self.all

    def read(self, handle):
        return self.mb.tag_get_data(self.tag_handle, handle, flat = True).astype(np.int64)

    @property
    def all_flagged_elements(self):
        return np.array(  list(self.flag.values())).astype(int)

    @property
    def all_flags(self):
        return np.array(list(self.flag.keys())).astype(int)

    @property
    def all(self):
        return self.read(self.elements_handle)

    @property
    def boundary(self):
        return self.read(self.boundary_elements.get_array())

    @property
    def internal(self):
        return self.read(self.internal_elements.get_array())


class MoabVariable(object):
    def __init__(self, core, name_tag, var_type="volumes", data_size=1, data_format="float", data_density="sparse",
                 entity_index=None):
        # pdb.set_trace()
        dic_dtf = {'float': types.MB_TYPE_DOUBLE, 'int': types.MB_TYPE_INTEGER, 'bool': types.MB_TYPE_BIT}
        dic_dst = {'dense': types.MB_TAG_DENSE, 'sparse': types.MB_TAG_SPARSE, 'bit': types.MB_TAG_BIT}
        dic_elem = {'nodes': core.all_nodes, 'edges': core.all_edges, 'faces': core.all_faces, 'volumes': core.all_volumes}
        self.mb = core.mb
        self.var_type = var_type
        self.data_format = data_format
        self.data_size = data_size
        self.data_density = data_density
        self.name_tag = name_tag
        self.custom = False
        if var_type in dic_elem:
            self.elements_handle = dic_elem[var_type]
        if entity_index is not None:
            self.elements_handle = self.range_index(entity_index)
            self.custom = True
        if data_density not in dic_dst:
            print("Please define a valid tag type")
        if data_format not in dic_dtf:
            print("Please define a valid data format")
        self.tag_handle = self.mb.tag_get_handle(name_tag, data_size, dic_dtf[data_format], dic_dst[data_density], True)
        print("Component class {0} successfully intialized".format(self.name_tag))

    def __call__(self):
        return self.mb.tag_get_data(self.tag_handle, self.elements_handle)

    def __setitem__(self, index, data):
        if isinstance(index, np.int64) or isinstance(index, int):
            el_handle = np.array([self.elements_handle[index]])
        elif not isinstance(index, np.ndarray) and index is not None:
            el_handle = self.elements_handle[index].get_array()
        else:
            el_handle = self.elements_handle.get_array(index)
        if isinstance(data, int) or isinstance(data, float) or isinstance(data, bool):
            data = data * np.ones((el_handle.size, self.data_size)).astype(self.data_format)
        elif (isinstance(data, np.ndarray)) and (len(data) == self.data_size):
            data = np.tile(data, (el_handle.size, 1)).astype(self.data_format)
        elif isinstance(data, list) & (len(data) == self.data_size):
            data = np.array(data)
            data = np.tile(data, (el_handle.size, 1)).astype(self.data_format)
        self.mb.tag_set_data(self.tag_handle, el_handle, data)

    def __getitem__(self, index):
        if isinstance(index, np.int64) or isinstance(index, int):
            el_handle = np.array([self.elements_handle[index]])
        elif not isinstance(index, np.ndarray) and index is not None:
            el_handle = self.elements_handle[index].get_array()
        else:
            el_handle = self.elements_handle.get_array(index)
        return self.mb.tag_get_data(self.tag_handle, el_handle)

    def __str__(self):
        string = "{0} variable: {1} based - Type: {2} - Length: {3} - Data Type: {4}"\
            .format(self.name_tag.capitalize(), self.var_type.capitalize(), self.data_format.capitalize(),
                    self.data_size, self.data_density.capitalize())
        if self.custom:
            string = string + " - Custom variable"
        return string

    def __len__(self):
        return len(self.elements_handle)



"""
Module for management of fine scale mesh
"""
import time
import pdb
import numpy as np
from preprocessor.meshHandle.meshComponents import MoabVariable
from preprocessor.meshHandle.configTools.configClass import variableInit
from math import sqrt
from pymoab import core, types, rng, topo_util
from . corePymoab import CoreMoab
from . meshComponents import MeshEntities, MoabVariable
import yaml

print('Standard fine-scale mesh loaded: No multiscale components available')


class FineScaleMesh:
    def __init__(self, mesh_file, dim=3, var_config=None):
        self.var_config = var_config
        self.dim = dim
        self.core = CoreMoab(mesh_file, dim)
        if mesh_file is not None:
            self.run()

    def run(self):
        self.init_entities()
        self.init_variables()
        self.macro_dim()
        self.init_dimmension()

    def init_variables(self):
        if self.var_config is None:
            self.var_config = variableInit()
        for command in self.var_config.get_var(self.core.level):
            exec(command)

    def init_entities(self):
        self.nodes = MeshEntities(self.core, entity_type = "nodes")
        self.edges = MeshEntities(self.core, entity_type="edges")
        self.faces = MeshEntities(self.core, entity_type = "faces")
        if self.dim == 3:
            self.volumes = MeshEntities(self.core, entity_type = "volumes")

    def __len__(self):
        if self.dim == 3:
            return len(self.volumes)
        elif self.dim ==2:
            return len(self.faces)

    def macro_dim(self):
        # coords = self.core.mb.get_coords(self.core.all_nodes).reshape(len(self.core.all_nodes),3)
        min_coord = self.nodes.coords[:].min(axis = 0)
        max_coord  = self.nodes.coords[:].max(axis = 0)
        self.rx = (min_coord[0], max_coord[0])
        self.ry = (min_coord[1], max_coord[1])
        self.rz= (min_coord[2], max_coord[2])

    def init_dimmension(self):
        # center = MoabVariable(self.core, data_size=3, var_type=0, data_format='float', name_tag='CENTER', data_density='dense')
        # import pdb; pdb.set_trace()
        pass

    def init_center(self):
        self.core.create_tag_handle('CENTER',3)
        #centro dos volumes
        centers = np.zeros((len(self.core.all_volumes),3)).astype('float')
        index = 0
        for volume in self.core.all_volumes:
            centers[index] = self.get_centroid(volume)
            index += 1
        self.core.set_data("CENTER",centers)
        #centro das faces
        centers = np.zeros((len(self.core.all_faces), 3)).astype('float')
        index = 0
        for face in self.core.all_faces:
            centers[index] = self.get_centroid(face)
            index += 1
        self.core.set_data("CENTER", centers, range_el = self.core.all_faces)

        #centro das arestas
        centers = np.zeros((len(self.core.all_edges), 3)).astype('float')
        index = 0
        for edge in self.core.all_edges:
            centers[index] = self.get_centroid(edge)
            index += 1
        self.core.set_data("CENTER", centers, range_el = self.core.all_edges)

    def init_normal(self):
        self.core.create_tag_handle('NORMAL', 3)
        normal = np.zeros((len(self.core.all_faces), 3)).astype('float')
        index = 0
        for face in self.core.all_faces:
            verts = self.core.mb.get_connectivity(face)
            coords = np.array([self.core.mb.get_coords([vert]) for vert in verts])
            vec1 = coords[1] - coords[0]
            vec2 = coords[2] - coords[0]
            cross = np.cross(vec1,vec2)
            normal[index] = cross/np.linalg.norm(cross)
            index += 1
        self.core.set_data("NORMAL", normal, range_el=self.core.all_faces)


    def init_volume(self):
        mat = np.zeros(self.core.all_volumes.size())
        index = 0
        for vol in  self.core.all_volumes:
            mat[index] = self.get_volume(vol)
            index += 1
        self.core.create_tag_handle("VOLUME",1)
        #pdb.set_trace()
        self.core.set_data("VOLUME",mat)
        #volTag = self.mb.tag_get_handle("VOLUME", 1, types.MB_TYPE_DOUBLE, types.MB_TAG_DENSE, True)
        #self.mb.tag_set_data(volTag, self.all_volumes, mat)
        #return volTag

    def get_centroid(self, entity):
        verts = self.core.mb.get_connectivity(entity)
        coords = np.array([self.core.mb.get_coords([vert]) for vert in verts])
        qtd_pts = len(verts)
        #print qtd_pts, 'qtd_pts'
        coords = np.reshape(coords, (qtd_pts, 3))
        pseudo_cent = sum(coords)/qtd_pts
        return pseudo_cent

    def get_volume(self,entity):
        #input: entity tag
        #ouput: volume of a entity
        verts = self.core.mb.get_connectivity(entity)
        coords = np.array([self.core.mb.get_coords([vert]) for vert in verts])
        qtd_pts = len(verts)
        if qtd_pts == 4:
            pass
            vect_1 = coords[1] - coords[0]
            vect_2 = coords[2] - coords[0]
            vect_3 = coords[3] - coords[0]
            vol_eval = abs(np.dot(np.cross(vect_1, vect_2), vect_3)) / 6.0
        elif qtd_pts == 8:
            pass
            #SEGUNDA ATIVIDADE PRA RENATINHA
            #CALCULAR O VOLUME DO HEXAEDRO IERREGULAR DADO OS 8 PONTOS

            # pass
            # #pdb.set_trace()
            # vect_1 = coords[7] - coords[0]
            # vect_2 = coords[1] - coords[0]
            # vect_3 = coords[3] - coords[5]
            # vect_4 = coords[4] - coords[6]
            # vect_5 = coords[5] - coords[0]
            # vect_6 = coords[2] - coords[0]
            # vect_7 = coords[6] - coords[3]
            # D1 = np.linalg.det(np.array([vect_1, vect_2, vect_3]))
            # D2 = np.linalg.det(np.array([vect_1, vect_4, vect_5]))
            # D3 = np.linalg.det(np.array([vect_1, vect_6, vect_7]))
            # vol_eval = ((D1+D2+D3)/2)
            vol_eval = 1
            return vol_eval
        else:
            vol_eval  = 0
        return vol_eval

    def get_tetra_volume(self, tet_nodes):
        vect_1 = tet_nodes[1] - tet_nodes[0]
        vect_2 = tet_nodes[2] - tet_nodes[0]
        vect_3 = tet_nodes[3] - tet_nodes[0]
        vol_eval = abs(np.dot(np.cross(vect_1, vect_2), vect_3))/1
        return vol_eval

    def get_piram_volume(self,tet_nodes):
        pass

    @staticmethod
    def point_distance(coords_1, coords_2):
        dist_vector = coords_1 - coords_2
        distance = sqrt(np.dot(dist_vector, dist_vector))
        return distance



"""
Schemes of partitioning
"""

#import pdb
import numpy as np
from numba import jit

#para definir funções de particionamento deve-se atentar as saidas
#é necessário retornar um vetor tag que associa uma partição a cada celula fina
#também é necessario retornar o uma matrix
# n x 3 com o centro de cada volume coarse, n -> numero de elementos do volume coarse

#a função tagAdjust remove renumera os elementos da malha coarse de maneira que eles permanecam
# continuos contando de 0 ao N de volumes coarse, a função tb remover os centros indesejados

#para adicionar novos esquemas basta criar um esquema com numeracao sequencial
#ex. scheme2  , mais shceme3
#a seção campo de leitura do arquivo msCoarse.ini deverá ter nome correspondente
#ex [Coarsening_2_Input] e [Coarsening_3_Input]
#todos atributos serao passados para funcao correspondente na ordem definida
#por default a leitura dos elementos é float, caso necessário. converta para int
#ex int(nx)

def scheme1(centerCoord, num_of_vol, rx,ry,rz ,nx = 3, ny = 3, nz =3 ):
    #input : centerCoord - > array with the center of elements
    #        num_of_vol = number of volumes

    #        rx,ry,rz - (min, max) values of x,y,z of the phyisical domain
    #        nx, ny, nz
    # msh -> objeto da clase meshUtil
    #centerCoord = msh.readData("CENTER")
    nx = int(nx)
    ny = int(ny)
    nz = int(nz)
    if (rz[1] == 0)  &  (rz[0] == 0):
        nz = 1
        rz = (-1,1)
    box = np.array([0, (rx[1] - rx[0])/nx, 0,(ry[1] - ry[0]) /ny, 0,(rz[1] - rz[0])/(nz+0)]).reshape(3,2)
    cent_coord_El1 = box.sum(axis =1)/2
    tag = np.zeros(num_of_vol).astype("int")
    coarseCenters = np.zeros((nx*ny*nz,3))
    index = 0
    init_coords = np.array([rx[0],ry[0],rz[0]])
    for x in range(nx):
        for y in range(ny):
            for z in range(nz):
                inc = np.multiply(box[:,1], np.array([x,y,z]))

                #cent = cent_coord_El1 + inc
                coarseCenters[index] = cent_coord_El1 + inc
                # pdb.set_trace()

                #inc = np.array([(nx) * x, (ny) * y, (nz) * z])
                boxMin = box[:,0] + inc + init_coords
                boxMax = box[:,1] + inc + init_coords
                point = checkinBox(centerCoord,x=(boxMin[0], boxMax[0]), y=(boxMin[1], boxMax[1]) , z=(boxMin[2], boxMax[2]))
                tag[point] = index
                index += 1
    return tagAdjust(tag,coarseCenters)


def scheme2(centerCoord, num_of_vol, rx,ry,rz ,nx = 3, ny = 3, nz =3 ):
    #input : centerCoord - > array with the center of elements
    #        num_of_vol = number of volumes

    #        rx,ry,rz - (min, max) values of x,y,z of the phyisical domain
    #        nx, ny, nz
    # msh -> objeto da clase meshUtil
    #centerCoord = msh.readData("CENTER")
    print("ESQUEMA TIPO 2 PARTICIONAMENTO")
    nx = int(nx)
    ny = int(ny)
    nz = int(nz)
    box = np.array([0, (rx[1] - rx[0])/nx, 0,(ry[1] - ry[0]) /ny, 0,(rz[1] - rz[0])/(nz+0)]).reshape(3,2)
    cent_coord_El1 = box.sum(axis =1)/2
    tag = np.zeros(num_of_vol).astype("int")
    coarseCenters = np.zeros((nx*ny*nz,3))
    index = 0

    for x in range(nx):
        for y in range(ny):
            for z in range(nz):
                inc = np.multiply(box[:,1], np.array([x,y,z]))
                #cent = cent_coord_El1 + inc
                coarseCenters[index] = cent_coord_El1 + inc
                #inc = np.array([(nx) * x, (ny) * y, (nz) * z])
                boxMin = box[:,0] + inc
                boxMax = box[:,1] + inc
                point = checkinBox(centerCoord,x=(boxMin[0], boxMax[0]), y=(boxMin[1], boxMax[1]) , z=(boxMin[2], boxMax[2]))
                tag[point] = index
                index += 1
    #return tagAdjust(tag,coarseCenters)


def scheme3(centerCoord, num_of_vol, rx,ry,rz ,nx = 3, ny = 3, nz =3 ):
    #input : centerCoord - > array with the center of elements
    #        num_of_vol = number of volumes

    #        rx,ry,rz - (min, max) values of x,y,z of the phyisical domain
    #        nx, ny, nz
    # msh -> objeto da clase meshUtil
    #centerCoord = msh.readData("CENTER")
    print("ESQUEMA 3")
    nx = int(nx)
    ny = int(ny)
    nz = int(nz)
    box = np.array([0, (rx[1] - rx[0])/nx, 0,(ry[1] - ry[0]) /ny, 0,(rz[1] - rz[0])/(nz+0)]).reshape(3,2)
    cent_coord_El1 = box.sum(axis =1)/2
    tag = np.zeros(num_of_vol).astype("int")
    coarseCenters = np.zeros((nx*ny*nz,3))
    index = 0
    for x in range(nx):
        for y in range(ny):
            for z in range(nz):
                inc = np.multiply(box[:,1], np.array([x,y,z]))
                #cent = cent_coord_El1 + inc
                coarseCenters[index] = cent_coord_El1 + inc
                #inc = np.array([(nx) * x, (ny) * y, (nz) * z])
                boxMin = box[:,0] + inc
                boxMax = box[:,1] + inc
                point = checkinBox(centerCoord,x=(boxMin[0], boxMax[0]), y=(boxMin[1], boxMax[1]) , z=(boxMin[2], boxMax[2]))
                tag[point] = index
                index += 1
    return tagAdjust(tag,coarseCenters)



#checa se um ponto esta dentro de um cubo

@jit(parallel=True)
def checkinBox(coords, x , y, z):
    tag1 = (coords[:,0] > x[0])   &  (coords[:,0] < x[1])
    tag2 = (coords[:,1] > y[0])   &  (coords[:,1] < y[1])
    tag3 = (coords[:,2] > z[0])   &  (coords[:,2] < z[1])
    return tag1 & tag2 & tag3



#função para corrigir os tags em caso que as malhas geradoras possuam volumes grossos sem celulas dentro
#e remover as respectivas coordenadas do centro dos volumes das malhas primais
#  @jit(parallel=True, cache=True)
def tagAdjust(tag, coarseCenter):
    # msh -> objeto da clase meshUtil
    fineTag =  tag
    elementsOriginal = [*set(tag)]
    elementsNovo = [*set(range(len(elementsOriginal)))]
    elementsMissing = set(range(len(coarseCenter))) - set(elementsOriginal)
    for elo, eln in zip(elementsOriginal,elementsNovo):
        if elo != eln:
            pointer = (tag == elo)
            fineTag[pointer] = eln
    return fineTag.astype(int) , np.delete(coarseCenter, [*elementsMissing], axis = 0)



"""
Module for implementation of multiscale mesh and CoarseVolumes objects functionalities
"""

import pdb
from . finescaleMesh import FineScaleMesh
from ..msCoarseningLib import algoritmo
from ..msCoarseningLib.partitionTools import partitionManager
from . meshComponents import MoabVariable
from . mscorePymoab import MsCoreMoab
from . meshComponentsMS import MoabVariableMS,  MeshEntitiesMS
from ..meshHandle.configTools.configClass import variableInit, coarseningInit
from pymoab import core, types, rng
import numpy as np



print('Initializing Finescale Mesh for Multiscale Methods')


class FineScaleMeshMS(FineScaleMesh):
    def __init__(self, mesh_file, dim=3, var_config=None):
        self.var_config = var_config
        super().__init__(mesh_file, dim)
        print("Creating Coarse Grid")
        self.coarse = MultiscaleCoarseGrid(self, var_config)
        self.enhance_entities()

    def __getitem__ (self, key):
        if not isinstance(key, int):
            raise ValueError("Invalid key type provided")
        return self.coarse.elements[key]

    def enhance_entities(self):
        for i,el in zip(range(len(self.coarse.elements)), self.coarse.elements):
            el(i,self.coarse)

    def init_entities(self):
        self.nodes = MeshEntitiesMS(self.core, entity_type="node")
        self.edges = MeshEntitiesMS(self.core, entity_type="edges")
        self.faces = MeshEntitiesMS(self.core, entity_type="faces")
        if self.dim == 3:
            self.volumes = MeshEntitiesMS(self.core, entity_type="volumes")

    def init_variables(self):
        if self.var_config is None:
            self.var_config = variableInit()
        for command in self.var_config.get_var(self.core.level):
            exec(command)

    def init_partition(self):
        coarse_config = coarseningInit()
        partitioner = partitionManager(self, coarse_config)
        [partition, coarse_center] = partitioner()
        if isinstance(partition, str) and partition == 'parallel':
            return self.init_partition_parallel()
        else:
            return partition

    def init_partition_parallel(self):
        if self.dim == 3:
            partition = MoabVariable(self.core, data_size=1,
                                     var_type="volumes", data_format="int",
                                     name_tag="Parallel", data_density="sparse")
        elif self.dim == 2:
            partition = MoabVariable(self.core, data_size=1, var_type="faces",
                                     data_format="int", name_tag="Parallel",
                                     data_density="sparse")
        return partition


class CoarseVolume(FineScaleMeshMS):
    def __init__(self, father_core, dim, i, coarse_vec, var_config=None):
        self.var_config = var_config
        self.dim = dim
        self.level = father_core.level + 1
        self.coarse_num = i
        self.coarse_vec = coarse_vec
        print("Level {0} - Volume {1}".format(self.level,self.coarse_num))
        self.core = MsCoreMoab(father_core, i, coarse_vec)
        self.init_entities()
        self.init_variables()
        self.init_coarse_variables()
        self.macro_dim()

    def init_variables(self):
        if self.var_config is None:
            self.var_config = variableInit()
        for command in self.var_config.get_var(self.core.level, self.coarse_num):
            exec(command)

    def __call__(self, i, general):
        self.nodes.enhance(i, general)
        self.edges.enhance(i, general)
        self.faces.enhance(i, general)
        if self.dim == 3:
            self.volumes.enhance(i, general)

    def init_coarse_variables(self):
        pass


class GetCoarseItem(object):
    def __init__(self, adj,tag, dic):
        self.fun = adj
        self.tag = tag
        self.dic = dic

    def __len__(self):
        return len(self.dic)

    def __call__(self, item):
        tmp = self.dic[item]
        el_list = rng.Range()
        for e in tmp:
            el_list.insert(e[0])
        return self.fun(self.tag, el_list)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.fun(self.tag, self.dic[item]).ravel()
        elif isinstance(item, slice):
            start = item.start
            step = item.step
            stop = item.stop
            if step == None:
                step = 1
            if start == None:
                start = 0
            if stop == None:
                stop = len(self.dic)
            array = np.array(range(start, stop, step))
            s = np.array([])
            for el in array:
                s = np.concatenate((s, self.__getitem__(int(el))))
            return s


class MultiscaleCoarseGrid(object):
    def __init__(self, M, var_config):
        self.mb = M.core.mb
        self.partition = M.init_partition()
        self.elements = [CoarseVolume(M.core, M.dim, i, self.partition[:].ravel() == i, var_config) for i in range(self.partition[:].max()+1 )]
        self.num_coarse = len(self.elements)
        self.num = {"nodes": 0, "node": 0, "edges": 1, "edge": 1, "faces": 2, "face": 2, "volumes": 3, "volume": 3,
                             0: 0, 1: 1, 2: 2, 3: 3}
        self.local_volumes_tag = [volume.core.handleDic[volume.core.id_name] for volume in self.elements]
        self.father_tag = M.core.handleDic[M.core.id_name]
        self.global_tag = M.core.handleDic["GLOBAL_ID"]
        self._all_volumes = M.core.all_volumes
        self._all_faces = M.core.all_faces
        self._all_edges = M.core.all_edges
        self._all_nodes = M.core.all_nodes
        self.find_coarse_neighbours()
        self.interfaces_faces = GetCoarseItem(self.mb.tag_get_data, self.father_tag, self._faces)
        self.interfaces_edges = GetCoarseItem(self.mb.tag_get_data, self.father_tag, self._edges)
        self.interfaces_nodes = GetCoarseItem(self.mb.tag_get_data, self.father_tag, self._nodes)
        self.iface_coarse_neighbors = self._internal_faces(M)

    def _internal_faces(self, M):
        #faces = np.array([self.mb.tag_get_data(self.father_tag,el[0]).ravel() for el in self._faces]).ravel()
        #faces = [self.interfaces_faces[int(el)] for el in range(len(self.interfaces_faces))]
        faces = []
        for el in range(len(self.interfaces_faces)):
            faces.append(self.interfaces_faces[el][0])
        partition = self.partition[:].ravel()
        internal = faces[0:self.num_internal_faces]
        external = faces[self.num_internal_faces:]
        internal_volumes = M.faces.bridge_adjacencies(internal, interface="faces",target="volumes")
        external_volumes = M.faces.bridge_adjacencies(external, interface="faces",target="volumes")
        int_neigh = np.vstack((partition[internal_volumes[:,0]],partition[internal_volumes[:,1]])).T
        ext_neigh = np.zeros((external_volumes.shape[0],2))
        ext_neigh[:,0], ext_neigh[:,1] = partition[external_volumes].ravel(), partition[external_volumes].ravel()
        return np.vstack((int_neigh,ext_neigh)).astype("uint64")

    def find_coarse_neighbours(self):
        self.connectivities = np.zeros((self.num_coarse,self.num_coarse+1 ,3)).astype('bool')
        self._nodes_neighbors  = np.zeros((self.num_coarse,self.num_coarse+1), dtype = object)
        self._edges_neighbors  = np.zeros((self.num_coarse,self.num_coarse+1), dtype = object)
        self._faces_neighbors  = np.zeros((self.num_coarse,self.num_coarse+1), dtype = object)
        self._nodes_neighbors[:], self._edges_neighbors[:], self._faces_neighbors[:] = None , None, None
        self._nodes = list()
        self._faces = list()
        self._edges = list()
        self.all_nodes_neighbors = rng.Range()
        self.all_edges_neighbors = rng.Range()
        self.all_faces_neighbors = rng.Range()
        self.all_volumes_neighbors = rng.Range()
        #print(self.all_nodes_neighbors, self.all_edges_neighbors, self.all_faces_neighbors)
        # import pdb; pdb.set_trace()
        node_count, edge_count, face_count = 0, 0, 0
        for x in range(self.num_coarse):
            for y in range(x+1,self.num_coarse):
                node_intersect = rng.intersect(self.elements[x].core.boundary_nodes, self.elements[y].core.boundary_nodes)
                if not node_intersect.empty():
                    self._nodes.append(node_intersect)
                    #self._nodes = np.append(self._nodes,node_intersect)
                    self._nodes_neighbors[x,y], self._nodes_neighbors[y,x],= node_count ,node_count
                    self.connectivities[x, y, 0],self.connectivities[y, x, 0] = True, True
                    node_count += 1
                    [self.all_nodes_neighbors.insert(e) for e in node_intersect]
                edges_intersect = rng.intersect(self.elements[x].core.boundary_edges, self.elements[y].core.boundary_edges)
                if not edges_intersect.empty():
                    self._edges.append(edges_intersect)
                    # self._edges = np.append(self._edges,edges_intersect)
                    self._edges_neighbors[x,y], self._edges_neighbors[y,x]= edge_count ,edge_count
                    self.connectivities[x, y, 1], self.connectivities[y, x, 1] =  True, True
                    edge_count += 1
                    [self.all_edges_neighbors.insert(e) for e in edges_intersect]
                faces_intersect = rng.intersect(self.elements[x].core.boundary_faces, self.elements[y].core.boundary_faces)
                if not faces_intersect.empty():
                    self._faces.append(faces_intersect)
                    #self._faces = np.append(self._faces,faces_intersect)
                    self._faces_neighbors[x,y], self._faces_neighbors[y,x]= face_count ,face_count
                    self.connectivities[x, y, 2],self.connectivities[y, x, 2]  = True, True
                    face_count += 1
                    [self.all_faces_neighbors.insert(e) for e in faces_intersect]
        self.num_internal_nodes = node_count
        self.num_internal_edges = edge_count
        self.num_internal_faces = face_count

        for x in range(self.num_coarse):
            #  fix the interesection - second variable poorly choosen
            node_intersect = rng.subtract(self.elements[x].core.boundary_nodes, self.all_nodes_neighbors)
            if not node_intersect.empty():
                self._nodes.append(node_intersect)
                self._nodes_neighbors[x, -1] = node_count
                self.connectivities[x, -1, 0] = True
                node_count += 1
            edge_intersect = rng.subtract(self.elements[x].core.boundary_edges, self.all_edges_neighbors)
            if not edge_intersect.empty():
                self._edges.append(edge_intersect)
                self._edges_neighbors[x, -1] = edge_count
                self.connectivities[x, -1, 1] = True
                edge_count += 1
            face_intersect = rng.subtract(self.elements[x].core.boundary_faces, self.all_faces_neighbors)
            if not face_intersect.empty():
                self._faces.append(face_intersect)
                self._faces_neighbors[x, -1] = face_count
                self.connectivities[x, -1, 2] = True
                face_count += 1

    def iface_neighbors(self, x):
        tmp = -1* np.ones(self._faces_neighbors[x].shape)
        tag = self._faces_neighbors[x] != None
        tmp[tag] = self._faces_neighbors[x,tag]
        #import pdb; pdb.set_trace()
        indices = np.where(tmp >= 0)[0]
        return indices, tmp[indices].astype(int)

    def iedge_neighbors(self, x):
        tmp = -1* np.ones(self._edges_neighbors[x].shape)
        tag = self._edges_neighbors[x] != None
        tmp[tag] = self._edges_neighbors[x,tag]
        #import pdb; pdb.set_trace()
        indices = np.where(tmp >= 0)[0]
        return indices, tmp[indices].astype(int)

    def inode_neighbors(self, x):
        tmp = -1* np.ones(self._nodes_neighbors[x].shape)
        tag = self._nodes_neighbors[x] != None
        tmp[tag] = self._nodes_neighbors[x,tag]
        #import pdb; pdb.set_trace()
        indices = np.where(tmp >= 0)[0]
        return indices, tmp[indices].astype(int)

    def father_to_local_id(self, vec_range,  element, target):
        flag = self.num[element]
        vec = self.create_range_vec(vec_range)
        if flag == 0:
            handle = self.range_index(vec, self._all_nodes)
        elif flag == 1:
            handle = self.range_index(vec, self._all_edges)
        elif flag == 2:
            handle = self.range_index(vec, self._all_faces)
        elif flag == 3:
            handle = self.range_index(vec, self._all_volumes)
        return self.mb.tag_get_data(self.local_volumes_tag[target],handle)

    def neighbours(self, x,y, element):
          flag = self.num[element]
          if flag == 0:
              return self.mb.tag_get_data(self.father_tag, self._nodes[self.nodes_neighbors[x,y]])
          elif flag == 1:
              return self.mb.tag_get_data(self.father_tag, self._edges[self.edges_neighbors[x,y]])
          elif flag == 2:
              return self.mb.tag_get_data(self.father_tag, self._faces[self.faces_neighbors[x,y]])

    @property
    def all_interface_nodes(self):
        return self.mb.tag_get_data(self.father_tag, self.all_nodes_neighbors)

    @property
    def all_interface_edges(self):
        return self.mb.tag_get_data(self.father_tag, self.all_edges_neighbors)

    @property
    def all_interface_faces(self):
        return self.mb.tag_get_data(self.father_tag, self.all_faces_neighbors)

    @property
    def all_neighbors_volumes(self):
        return self.mb.tag_get_data(self.father_tag, self.all_volumes_neighbors)

    def create_range_vec(self, index):
        range_vec = None
        if isinstance(index, int) or isinstance(index, np.integer):
            range_vec = np.array([index]).astype("uint")
        elif isinstance(index, np.ndarray):
            if index.dtype == "bool":
                range_vec = np.where(index)[0]
            else:
                range_vec = index
        elif isinstance(index, slice):
            start = index.start
            stop = index.stop
            step = index.step
            if start is None:
                start = 0
            if stop is None:
                stop = len(self)
            if step is None:
                step = 1
            if start < 0:
                start = len(self) + start + 1
            if stop < 0:
                stop = len(self) + stop + 1
            range_vec = np.arange(start, stop, step).astype('uint')
        elif isinstance(index, list):
            range_vec = np.array(index)
        return range_vec

    def read_data(self, name_tag, index_vec = np.array([]), range_el = None):
        if range_el is None:
            range_el = self.all_volumes
        if index_vec.size > 0:
            range_el = self.range_index(index_vec,range_el)
        try:
            handle_tag = self.handleDic[name_tag]
            return self.mb.tag_get_data(handle_tag, range_el)
        except KeyError:
            print("Tag not found")

    def range_index(self, vec_index, range_handle = None):
        if range_handle is None:
            range_handle = self.all_volumes
        if vec_index.dtype == "bool":
            vec = np.where(vec_index)[0]
        else:
            vec = vec_index.astype("uint")
        handles = np.asarray(range_handle)[vec.astype("uint")].astype("uint")
        return rng.Range(handles)



from scipy.spatial.qhull import Delaunay
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import tools.view as tl
import networkx as nx
from scipy.spatial import ConvexHull
x = 22

G = nx.graph.Graph()
#finding neighbors of the x element
neighbors = np.setdiff1d(M.coarse.elements[x].nodes.coarse_neighbors, len(M.coarse.elements))
neighbors = np.sort(np.insert(neighbors,1,x)) # XXX: ))
#neighbors = np.append(neighbors, x).astype(int)
#center coordinates of the CV
ccoord = M.volumes.center[l.coarse_center[x]]
#center coordinates of the cv coarse volume neigbhors
dcoords = M.volumes.center[l.coarse_center[np.setdiff1d(neighbors,x)]]
faces_neighbors = np.isin(M.coarse.partition[:].ravel()[l.interface_vol], neighbors)
all_neighbors_index = np.logical_and(faces_neighbors[:,0], faces_neighbors[:,1])
internal_tag = np.zeros(all_neighbors_index.shape, dtype="bool")
internal_tag[0:M.coarse.num_internal_faces] = 1

# removing bad tetrahedron that do not have the center
#modified_simplices = trian.simplices[np.sum(trian.simplices == trian.simplices.max(),axis=1) == 1]



#coordinates of the interfaces of the CV with its neighbors
x_tag = np.sum((M.coarse.partition[:].ravel()[l.interface_vol] == x),axis=1) >=1
cv_interfaces = np.logical_and(all_neighbors_index,x_tag)
cv_icoords = M.faces.center[l.center_face[np.logical_and(cv_interfaces, internal_tag)]]
#coordinates of the interface of the CV with the boundary
cv_xcoords = M.faces.center[l.center_face[np.logical_and(cv_interfaces, ~internal_tag)]]
#cv_neighbors = np.sum(np.isin(M.coarse.partition[:].ravel()[l.interface_vol],x),axis=1) >= 1


#coordinates of the interfaces of the CV neighbors among each other
icoords = M.faces.center[l.center_face[np.logical_and(~cv_interfaces, internal_tag)]]
#coordinates of the interfaces of the CV neighbors on the boundary
xcoords = M.faces.center[l.center_face[np.logical_and(~cv_interfaces, ~internal_tag)]]

#
# cv_interfaces = np.logical_and(all_neighbors_index, cv_neighbors)
# neighbors_interfaces = np.logical_and(all_neighbors_index, ~cv_neighbors)
#
# np.sum(np.isin(M.coarse.partition[:].ravel()[l.interface_vol],x),axis=1) == 1
#
# #coordinates of the interfaces of the CV with its neighbors
# cv_icoords = M.faces.center[l.center_face[np.logical_and(cv_interfaces, internal_tag)]]
# #coordinates of the interface of the CV with the boundary
# cv_xcoords = M.faces.center[l.center_face[np.logical_and(cv_interfaces, ~internal_tag)]]
#
# #coordinates of the interfaces of the CV neighbors among each other
# icoords = M.faces.center[l.center_face[np.logical_and(neighbors_interfaces, internal_tag)]]
# #coordinates of the interfaces of the CV neighbors on the boundary
# xcoords = M.faces.center[l.center_face[np.logical_and(neighbors_interfaces, ~internal_tag)]]

tcoords = np.vstack((dcoords, cv_icoords, cv_xcoords, icoords, xcoords, ccoord))
point_type = np.concatenate((0*np.ones(dcoords.shape[0]),1*np.ones(cv_icoords.shape[0]), 2*np.ones(cv_xcoords.shape[0]) , 3*np.ones(icoords.shape[0]), 4*np.ones(xcoords.shape[0]), 5*np.ones(ccoord.shape[0])))


tcoords = np.vstack((dcoords, cv_icoords, cv_xcoords, icoords, ccoord))
point_type = np.concatenate((0*np.ones(dcoords.shape[0]),1*np.ones(cv_icoords.shape[0]), 2*np.ones(cv_xcoords.shape[0]) , 3*np.ones(icoords.shape[0]),  5*np.ones(ccoord.shape[0])))


tcoords = np.vstack((dcoords, ccoord))
point_type = np.concatenate((0*np.ones(dcoords.shape[0]), 5*np.ones(ccoord.shape[0])))





trian = Delaunay(tcoords)

df = pd.DataFrame(point_type[trian.simplices])
df.to_excel("delaunay.xlsx", index=False)
modified_simplices = trian.simplices[np.sum(trian.simplices == trian.simplices.max(),axis=1) == 1]
a = tl.DelaunaySingle(trian.points, trian.simplices, point_type, 2)
#a = tl.DelaunaySingle(trian.points, trian.simplices,point_type,2)
# import pdb; pdb.set_trace()
#ncoords = np.vstack((dcoords, extra_coords_internal, extra_coords_external))

# a = tl.DelaunaySingle(, good_tetra, neighbors, x)
# # #perfoming delaunay triangulation
# # trian = Delaunay(dcoords)



"""
Geometric methods to compute volumes, areas, distances of the mesh entities
"""
# Geoemtric Module
# Create by Artur Castiel and Renata Tavares

#import pdb
import numpy as np
from numba import jit

# class for volume related problem
    #
    # def init_normal(self):
    #     self.core.create_tag_handle('NORMAL', 3)
    #     normal = np.zeros((len(self.core.all_faces), 3)).astype('float')
    #     index = 0
    #     for face in self.core.all_faces:
    #         verts = self.core.mb.get_connectivity(face)
    #         coords = np.array([self.core.mb.get_coords([vert]) for vert in verts])
    #         vec1 = coords[1] - coords[0]
    #         vec2 = coords[2] - coords[0]
    #         cross = np.cross(vec1,vec2)
    #         normal[index] = cross/np.linalg.norm(cross)
    #         index += 1
    #     self.core.set_data("NORMAL", normal, range_el=self.core.all_faces)
#@jit(parallel = True)

@jit(parallel = True)
def normal_vec_2d(coords0,coords1):
    vec = coords1 - coords0
    norm = np.linalg.norm(vec, axis = 1)
    norm = 1/norm
    return np.array([vec[:,1], -vec[:,0], vec[:,2] ]).T * norm[:,np.newaxis]
    # distance = (np.inner(vec, vec, axis = 0))


# @jit(nopython=True)
# def cross_numba(vec1,vec2):
#     vec1 = double(vec1)
#     vec2 = double(vec2)
#     result = np.zeros((vec1.shape[0],3))
#     result[:,0] = vec1[:,1]*vec2[:,2] - vec1[:,2]*vec2[:,1]
#     result[:,1] = vec1[:,2]*vec2[:,0] - vec1[:,0]*vec2[:,2]
#     result[:,2] = vec1[:,0]*vec2[:,1] - vec1[:,1]*vec2[:,0]
#     return result
# @jit
# def cross_numba(vec1, vec2):
#     """ Calculate the cross product of two 3d vectors. """
#     result = np.zeros((vec1.shape[0],3)
#     # [a1, a2, a3] = double(vec1[0]), double(vec1[1]), double(vec1[2])
#     # [b1, b2, b3] = double(vec2[0]), double(vec2[1]), double(vec2[2])
#     # result[0] = a2 * b3 - a3 * b2
#     # result[1] = a3 * b1 - a1 * b3
#     # result[2] = a1 * b2 - a2 * b1
# return result
@jit
def normal_vec(coords0, coords1, coords2):
    vec1 = coords1 - coords0
    vec2 = coords2 - coords0
    #cross = cross_numba(vec1,vec2)
    cross = np.cross(vec1,vec2)
    norm = np.linalg.norm(cross, axis = 1)
    norm = 1/norm
    return  cross * norm[:,np.newaxis]
    # a = cross * norm

def point_distance(coords_1, coords_2):
    dist_vector = coords_1 - coords_2
    distance = sqrt(np.dot(dist_vector, dist_vector))
    return distance

#@jit(parallel=True)
@jit
def get_average(coords_list):
    N = len(coords_list)
    return sum(coords_list)*(1/N)
#
# def tetraVolume(tet_nodes):
#     #input:
#     # A Matrix with 4x3 elements in which
#     # each line is one of the 4 nodes that
#     # a given tetrahedron is comprised
#     #ouput:
#     # the volume of the given tetrahedron
#     vect_1 = tet_nodes[1] - tet_nodes[0]
#     vect_2 = tet_nodes[2] - tet_nodes[0]
#     vect_3 = tet_nodes[3] - tet_nodes[0]
#     vol_eval = abs(np.dot(np.cross(vect_1, vect_2), vect_3))/6
#     return vol_eval
#
# def piramidVolume(pi_nodes):
#     #     P5           P4 _____ P3
#     #     /\             |     |
#     #    /  \            |     |
#     #   /____\           |_____|
#     # P1/4  P2/3        P1     P2
#     #input:
#     # A Matrix with 4x3 elements in which
#     # each line is one of the 5 nodes that
#     # a given piramid is comprised
#     # The 3 first nodes lie on the plane of the base (coplanar points) and must be connected.
#     # The fourth node of the matrix must be the top point (P5).
#     # ouput: the volume of the given piramid
#     vect_1 = pi_nodes[1] - pi_nodes[0]
#     vect_2 = pi_nodes[2] - pi_nodes[0]
#     base_area = abs(np.dot(vect_1, vect_2))
#
#     vect_3 = pi_nodes[3] - pi_nodes[0]
#     normal_vect = np.cross(vect_1, vect_2)
#     piram_height = np.dot(vect_3, normal_vect)/(np.linalg.norm(normal_vect)
#
#     piram_vol = (1/3)*base_area*piram_height
#     return(piram_vol)
#
# def hexahedronVolume(pi_nodes):
#     #
#     #    ______   <- F2
#     #   /     /|
#     #  /_____/ |
#     #  |     | |
#     #  | F1  | /
#     #  |_____|/
#     #
#     #
#     # P1 _____P2     P5  ____ P6
#     #   |     |         |    |
#     #   | F1  |         | F2 |
#     #   |_____|         |____|
#     # P4      P3     P8       P7
#     #
#     # F1 - Front Face
#     # F2 - Back Face
#     #  NOTE:
#     #  The given hexahedron may be irregular
#     #  The sketch above describes the connectivities
#     #
#     #input:
#     # A Matrix with 8x3 elements in which
#     # each line is one of the 8 nodes that
#     # a given hexahedron
#     #ouput:
#     # the volume of the given piramid
#     print(pi_nodes)
# def teste():
#     print("Entrou")
#     pass



"""
Use of Pymoab methods to manage the multiscale mesh
"""
#import pdb
from . corePymoab import CoreMoab
from pymoab import core, types, rng, topo_util
from pymoab import skinner as sk
import numpy as np
import time


class MsCoreMoab(CoreMoab):
    def __init__(self, father_core, num, coarse_vec):
        self.father_core = father_core
        self.dimension = father_core.dimension
        self.mb = father_core.mb
        self.level = father_core.level + 1
        self.coarse_num = num
        self.father_root_set = father_core.root_set
        self.root_set = self.mb.create_meshset(types.MESHSET_TRACK_OWNER) #types.MESHSET_TRACK_OWNER)
        self.mtu = father_core.mtu
        self.handleDic = father_core.handleDic
        if self.dimension == 3:
            self.all_volumes = father_core.all_volumes[coarse_vec]
            self.all_faces = self.mb.get_adjacencies(self.all_volumes, 2, False, op_type=types.UNION)
            self.all_edges = self.mb.get_adjacencies(self.all_volumes, 1, False, op_type=types.UNION)
            self.all_nodes = self.mb.get_adjacencies(self.all_volumes, 0, False, op_type=types.UNION)
        elif self.dimension == 2:
            self.all_volumes = rng.Range()
            self.all_faces = father_core.all_faces[coarse_vec]
            self.all_edges = self.access_handle(self.all_faces)
            self.all_nodes = rng.Range(self.mb.get_connectivity(self.all_faces))
        self.mb.add_entities(self.root_set, self.all_volumes)
        self.mb.add_entities(self.root_set, self.all_faces)
        self.mb.add_entities(self.root_set, self.all_edges)
        self.mb.add_entities(self.root_set, self.all_nodes)
        all_entities = self.mb.get_entities_by_handle(self.root_set)

        [self.boundary_nodes, self.boundary_edges, self.boundary_faces, self.boundary_volumes] = self.skinner_operation()
        self.internal_nodes = rng.subtract(self.all_nodes, self.boundary_nodes)
        self.internal_edges = rng.subtract(self.all_edges, self.boundary_edges)
        self.internal_faces = rng.subtract(self.all_faces, self.boundary_faces)
        self.internal_volumes = rng.subtract(self.all_volumes, self.boundary_volumes)
        if self.level == 1:
            self.id_name = "LOCAL_ID_L" + str(self.level) + "-" + str(self.coarse_num)
            self.father_id_name = self.father_core.id_name
        else:
            self.father_id_name = self.father_core.id_name
            self.id_name = self.father_id_name + str("L") + str(self.level) + "-" + str(self.coarse_num)


        self.init_id()
        # all_entities = self.mb.get_entities_by_handle(self.root_set)
        # print(all_entities)
        self.flag_dic = {key:[rng.intersect(all_entities,el) for el in value] for (key, value) in father_core.flag_dic.items()}

    def skinner_operation(self):
        self.skin = sk.Skinner(self.mb)
        if self.dimension == 3:
            faces_on_skin_handles  = self.skin.find_skin(0, self.all_volumes)
            edges_on_skin_handles = self.mtu.get_bridge_adjacencies(faces_on_skin_handles, 2, 1)
            nodes_on_skin_handles = self.mtu.get_bridge_adjacencies(faces_on_skin_handles, 2, 0)
            volumes_on_skin_handles = rng.intersect(self.mtu.get_bridge_adjacencies(faces_on_skin_handles, 0, 3), self.all_volumes)
        elif self.dimension == 2:
            edges_on_skin_handles = self.skin.find_skin(0, self.all_faces)
            nodes_on_skin_handles = self.mtu.get_bridge_adjacencies(edges_on_skin_handles, 1, 0)
            faces_on_skin_handles = rng.intersect(self.mtu.get_bridge_adjacencies(edges_on_skin_handles, 0, 2), self.all_faces)
            volumes_on_skin_handles = self.mtu.get_bridge_adjacencies(edges_on_skin_handles, 0, 3) #empty
        return [nodes_on_skin_handles, edges_on_skin_handles, faces_on_skin_handles, volumes_on_skin_handles]



from pymoab import core, types, rng, topo_util
from pymoab import skinner as sk
import numpy as np


def volume(p1, p2, p3, p4):
    d1 = p2 - p1
    d2 = p3 - p1
    d3 = p4 - p1
    matrix = np.vstack((d1, d2, d3))
    return np.abs((1/6)*np.linalg.det(matrix))


def volume_p(tet):
    vol = np.zeros(tet.shape[0])
    #import pdb; pdb.set_trace()
    for i in range(tet.shape[0]):
        p1, p2, p3, p4 = tet[i][0], tet[i][1], tet[i][2], tet[i][3]
        # d1 = p2 - p1
        # d2 = p3 - p1
        # d3 = p4 - p1
        matrix = np.hstack((np.vstack((p1, p2, p3, p4)), np.ones((4, 1))))
        vol[i] = (1/6)*np.linalg.det(matrix)
    return np.abs(vol)
    #matrix = np.vstack((d1, d2, d3))
    #return (1/6)*np.linalg.det(matrix)


class DelaunaySingle(object):
    def __init__(self, coords, elements, point_type, center):
        self.mb = core.Core()
        #self.point_type = np.concatenate((np.array([-1]), point_type))
        self.point_type = point_type
        verts = self.mb.create_vertices(coords)
        self.rs = self.mb.get_root_set()

        elements = elements + np.ones(elements.shape)
        self.tag_handle = self.mb.tag_get_handle("Teste", 1, types.MB_TYPE_INTEGER, types.MB_TAG_DENSE, create_if_missing = True)
        self.tag_node = self.mb.tag_get_handle("Nodes", 1, types.MB_TYPE_INTEGER, types.MB_TAG_DENSE, create_if_missing = True)

        for el in elements:
            tetra = self.mb.create_element(types.MBTET, el.ravel().astype("uint64"))

        # self.mtu = topo_util.MeshTopoUtil(self.mb)
        # self.mtu.construct_aentities(verts)
        #self.mtu.construct_aentities(verts)
        #self.elements = self.mb.get_entities_by_dimension(0, 3)
        self.mtu = topo_util.MeshTopoUtil(self.mb)
        self.all_elements = self.mb.get_entities_by_dimension(0, 3)
        self.nodes = self.mb.get_entities_by_dimension(0, 0)
        self.faces = self.skinner(self.all_elements)
        self.boundary_nodes = self.find_boundary_nodes()
        self.boundary_elements = self.find_boundary_elements()

        self.desired_bnodes = np.array(self.nodes)[self.point_type == 0]
        self.find_simple_bad()
        #self.select(self.boundary_nodes)
        #self.remove_vol(self.all_elements[0:120])
        #self.find_bad_vol()
        #import pdb; pdb.set_trace()
        #self.smooth()
        #import pdb; pdb.set_trace()
        self.create_vtk()

    def find_missing_nodes(self):
        return self.desired_bnodes[~np.isin(self.desired_bnodes,np.array(self.boundary_nodes))]

    def bad_vol(self):
        index = 0
        point_type = np.concatenate((np.array([-1]), self.point_type))
        while not self.condition():
            #import pdb; pdb.set_trace()
            index = index + 1
            connect = self.connectivities(self.boundary_elements, 4)
            flag = point_type[connect]
            all_tetra_ref = np.arange(connect.shape[0])
            num_tetra_bnodes = np.sum(np.isin(connect,np.array(self.boundary_nodes)), axis=1)
            four_bnodes_tetra = (num_tetra_bnodes == 4)
            three_bnodes_tetra = (num_tetra_bnodes == 3)


            # treating tetrahedron with 4 nodes on the boundary those tetrahedron
            # will be removed as no node are exclusive to these entities
            rm1 = all_tetra_ref[four_bnodes_tetra][~np.any(np.logical_and(connect[four_bnodes_tetra],flag[four_bnodes_tetra] == 0),axis=1)]
            #~np.any(np.logical_and(connect[four_bnodes_tetra],flag[four_bnodes_tetra] == 0),axis=1)
            #import pdb; pdb.set_trace()

            # treating tetrahedron with 3 nodes on the boundary those tetrahedron
            # will be removed as no node are exclusive to these entities
            rmp = all_tetra_ref[three_bnodes_tetra]
            # ~np.isin(connect[three_bnodes_tetra],np.array(self.boundary_nodes))
            tag = np.any(flag[three_bnodes_tetra] == 0,  axis=1)
            rm2 = rmp[~tag]
            tetra_tag = rmp[tag]
            rm3 = tetra_tag[np.all(~np.isin(connect[tetra_tag],np.array(self.boundary_nodes)) == (flag[tetra_tag] == 0), axis=1)]
            # flag[three_bnodes_tetra]
            # rm2 = all_tetra_ref[four_bnodes_tetra]
            remove_tetra = np.sort(np.hstack((rm1,rm2,rm3)))
            remove_flag = np.isin(all_tetra_ref, remove_tetra)

            self.remove_vol(self.select(self.boundary_elements,remove_flag))

            self.create_vtk()
            #import pdb; pdb.set_trace()

    def find_simple_bad(self):
        index = 0
        point_type = np.concatenate((np.array([-1]), self.point_type))
        missing = self.find_missing_nodes()
        coord = self.mb.get_coords(self.nodes).reshape(-1,3)]
        remov = rng.Range()
        while not self.condition():
            index = index + 1
            connect = self.connectivities(self.boundary_elements, 4)
            all_tetra_ref = np.arange(connect.shape[0])
            flag = point_type[connect]
            for el in missing:
                tetra_analize = np.any(np.isin(connect, el), axis=1)
                np.unique(connect[tetra_analize])
                tet_coord = (connect[tetra_analize] - np.ones(connect[tetra_analize].shape)).astype("uint64")
                #tet_coord1 = (connect - np.ones(connect.shape)).astype("uint64")
                vol_tetra = volume_p(coord[tet_coord])
                removable = self.select(self.boundary_elements, all_tetra_ref[tetra_analize][vol_tetra == vol_tetra.min()][0])
                remov = rng.unite(remov,removable)

            self.create_vtk()
            import pdb; pdb.set_trace()
            self.remove_vol(removable)

    def find_bad_vol(self):
        index = 0
        point_type = np.concatenate((np.array([-1]), self.point_type))
        while not self.condition():
            #import pdb; pdb.set_trace()
            index = index + 1
            connect = self.connectivities(self.boundary_elements, 4)
            flag = point_type[connect]
            np.all(flag == 3,axis=1)



            import pdb; pdb.set_trace()

            all_tetra_ref = np.arange(connect.shape[0])
            num_tetra_bnodes = np.sum(np.isin(connect,np.array(self.boundary_nodes)), axis=1)
            four_bnodes_tetra = (num_tetra_bnodes == 4)
            three_bnodes_tetra = (num_tetra_bnodes == 3)
            # treating tetrahedron with 4 nodes on the boundary those tetrahedron
            # will be removed as no node are exclusive to these entities
            rm1 = all_tetra_ref[four_bnodes_tetra][~np.any(np.logical_and(connect[four_bnodes_tetra],flag[four_bnodes_tetra] == 0),axis=1)]
            #~np.any(np.logical_and(connect[four_bnodes_tetra],flag[four_bnodes_tetra] == 0),axis=1)
            #import pdb; pdb.set_trace()

            # treating tetrahedron with 3 nodes on the boundary those tetrahedron
            # will be removed as no node are exclusive to these entities
            rmp = all_tetra_ref[three_bnodes_tetra]
            # ~np.isin(connect[three_bnodes_tetra],np.array(self.boundary_nodes))
            tag = np.any(flag[three_bnodes_tetra] == 0,  axis=1)
            rm2 = rmp[~tag]
            tetra_tag = rmp[tag]
            rm3 = tetra_tag[np.all(~np.isin(connect[tetra_tag],np.array(self.boundary_nodes)) == (flag[tetra_tag] == 0), axis=1)]
            # flag[three_bnodes_tetra]
            # rm2 = all_tetra_ref[four_bnodes_tetra]
            remove_tetra = np.sort(np.hstack((rm1,rm2,rm3)))
            remove_flag = np.isin(all_tetra_ref, remove_tetra)

            self.remove_vol(self.select(self.boundary_elements,remove_flag))

            self.create_vtk()
            #import pdb; pdb.set_trace()

    def smooth(self):
        point_type = np.concatenate((np.array([-1]), self.point_type))
        connect = self.connectivities(self.boundary_elements, 4)
        all_tetra_ref = np.arange(connect.shape[0])
        flag = point_type[connect]
        self.remove_vol(self.select(self.boundary_elements,np.all(flag == 3,axis=1)))

    def condition(self):
        #desired_bnodes = np.array(self.nodes)[self.point_type == 0]
        print("desired nodes:", self.desired_bnodes)
        print("boundary nodes:", np.array(self.boundary_nodes))
        print("nodes on boundary", self.desired_bnodes[np.isin(self.desired_bnodes,np.array(self.boundary_nodes))])
        print("missing", self.desired_bnodes[~np.isin(self.desired_bnodes,np.array(self.boundary_nodes))])
        #print(~desired_bnodes[np.isin(desired_bnodes,np.array(self.boundary_nodes))])
        return np.all(np.isin(self.desired_bnodes,np.array(self.boundary_nodes)))

    def select(self, range, vec):
        return rng.Range(np.array(range)[vec])

    def find_boundary_nodes(self):
        return self.adjs(self.faces, 0)

    def find_boundary_elements(self):
        #self.mtu = topo_util.MeshTopoUtil(self.mb)
        rnb = rng.Range()
        it = ([self.mtu.get_bridge_adjacencies(el, 2,3) for el in self.faces])

        for el in it:
            rnb = rng.unite(rnb, el)
        return rnb


    def create_vtk(self):
        for index,el in enumerate(self.all_elements):
            self.mb.tag_set_data(self.tag_handle, el, index)
        for index,el in enumerate(self.nodes):
            mp = self.point_type[index].astype(int)
            # import pdb; pdb.set_trace()
            self.mb.tag_set_data(self.tag_node, el, mp)
        meshset = self.mb.create_meshset()
        self.mb.add_entity(meshset, self.nodes)
        self.mb.write_file("delaunay0.vtk", [meshset])
        self.mb.write_file("delaunay1.vtk")

    def remove_vol(self, element):
        tetra_remove = rng.intersect(self.all_elements, element)
        tetra_remaining = rng.subtract(self.all_elements, tetra_remove)
        #import pdb; pdb.set_trace()
        faces = rng.subtract(self.adjs(tetra_remove, 2), self.adjs(tetra_remaining,2))
        edges = rng.subtract(self.adjs(tetra_remove, 1), self.adjs(tetra_remaining,1))
        nodes = rng.subtract(self.adjs(tetra_remove, 0), self.adjs(tetra_remaining,0))
        self.mb.delete_entity(tetra_remove)
        self.mb.delete_entity(faces)
        self.mb.delete_entity(edges)
        self.mb.delete_entity(nodes)
        self.all_elements = self.mb.get_entities_by_dimension(0, 3)
        self.nodes = self.mb.get_entities_by_dimension(0, 0)
        self.faces = self.skinner(self.all_elements)
        self.boundary_nodes = self.find_boundary_nodes()
        self.boundary_elements = self.find_boundary_elements()

    def adjs(self, range,x):
        rnb = rng.Range()
        #import pdb; pdb.set_trace()
        it = ([self.mb.get_adjacencies(el, x) for el in range])
        for el in it:
            rnb = rng.unite(rnb, el)
        return rnb

    def connectivities(self,range, i):
        return self.mb.get_connectivity(range).reshape((-1, i))

    def external_nodes(self):
        range = self.faces
        return np.unique(np.array(self.mb.get_connectivity(range)))

    def skinner(self, range):
        skin = sk.Skinner(self.mb)
        return skin.find_skin(self.rs,range)

class DelaunayView(object):
    def __init__(self, coords, elements, neighbors, center):
        self.mb = core.Core()
        self.mtu = topo_util.MeshTopoUtil(self.mb)
        skin = sk.Skinner(self.mb)
        verts = self.mb.create_vertices(coords)
        import pdb; pdb.set_trace()
        rs = self.mb.get_root_set()
        #import pdb; pdb.set_trace()
        #import pdb; pdb.set_trace()
        import pdb; pdb.set_trace()
        elements = elements + np.ones(elements.shape)
        tag_handle = self.mb.tag_get_handle("Teste", 1, types.MB_TYPE_INTEGER, types.MB_TAG_DENSE, create_if_missing = True)
        for el in elements:
            tetra = self.mb.create_element(types.MBTET, el.ravel().astype("uint64"))
        #self.mtu.construct_aentities(verts)
        elements = self.mb.get_entities_by_dimension(0, 3)
        for index,el in enumerate(elements):
            self.mb.tag_set_data(tag_handle,el,index)
        self.mb.write_file("delaunayVIEW.vtk")

class Delaunay(object):
    """Class for Visualization using moab of a Delaunay triangulation using SciPy."""

    def __init__(self, coords, elements, neighbors, center):
        #super(DelaunayView, self).__init__()
        self.mb = core.Core()
        self.mtu = topo_util.MeshTopoUtil(self.mb)
        skin = sk.Skinner(self.mb)
        verts = self.mb.create_vertices(coords)
        rs = self.mb.get_root_set()
        #import pdb; pdb.set_trace()
        #import pdb; pdb.set_trace()
        elements = elements + np.ones(elements.shape)

        tag_handle = self.mb.tag_get_handle("Teste", 1, types.MB_TYPE_INTEGER, types.MB_TAG_DENSE, create_if_missing = True)


        for el in elements:
            tetra = self.mb.create_element(types.MBTET, el.ravel().astype("uint64"))


        #self.mtu.construct_aentities(verts)
        elements = self.mb.get_entities_by_dimension(0, 3)

        for index,el in enumerate(elements):
            self.mb.tag_set_data(tag_handle,el,index)
        #import pdb; pdb.set_trace()
        bfaces = skin.find_skin(0, elements)
        print(bfaces)
        adj = np.array([(self.mb.get_connectivity(bface)) for bface in bfaces])
        adj = adj.reshape((len(adj), 3)).astype(int) - np.ones(adj.shape)
        missing = np.setdiff1d(np.where(~np.isin(neighbors,adj)), center)
        import pdb; pdb.set_trace()
        # finding boundary tetrahedron
        self.mb.write_file("delaunay01.vtk")
        emp = rng.Range()
        boundary_tetrahedron = rng.Range()
        for el in elements:
            boundary_intersect = rng.intersect(bfaces, self.mb.get_adjacencies(el, 2))
            if boundary_intersect is not emp:
                boundary_tetrahedron = rng.unite(boundary_tetrahedron, rng.Range(el))


        # for el in elements:
        #     #import pdb; pdb.set_trace()
        #     local = rng.Range(el)
        #     boundary_intersect = rng.intersect(bfaces, self.mb.get_adjacencies(local, 2))
        #     if boundary_intersect is not emp:
        #         # import pdb; pdb.set_trace()
        #         face_con = self.mb.get_adjacencies(boundary_intersect,0)
        #         el_con = self.mb.get_adjacencies(local, 0)
        #             #import pdb; pdb.set_trace()
        #         inside_node = int(rng.subtract(el_con, face_con)[0] - 1)
        #
        #         #inside_node = int(np.setdiff1d(el_con, face_con)[0] - 1)
        #         #import pdb; pdb.set_trace()
        #         #check if inside node is missing
        #         is_missing = bool(np.isin(inside_node, missing))
        #         if is_missing is True:
        #             boundary_tetrahedron = rng.unite(boundary_tetrahedron, local)
                #import pdb; pdb.set_trace()


        for el in boundary_tetrahedron:
            self.mb.tag_set_data(tag_handle,el,2)
            print(self.mb.get_connectivity(el))
        #print(boundary_tetrahedron)
        import pdb; pdb.set_trace()

        #self.mb.delete_entity(boundary_tetrahedron)
        print(boundary_tetrahedron)
        self.mb.write_file("delaunay02.vtk")



import setuptools
from setuptools.command.install import install

with open("README.md", "r") as fh:
    long_description = fh.read()
with open("requirements.txt") as fr:
    install_requires = fr.read().splitlines()

setuptools.setup(
    name="impress-padmec",
    version="1.5.0",
    author="Artur Castiel, Gabriel Mendes",
    description="Intuitive Multilevel Preprocessor for Smart Simulation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires = install_requires,
    url="https://github.com/padmec-reservoir/impress",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
)



"""
Generator of multiscale mesh entities and tags
"""
import numpy as np
from . meshComponents import MoabVariable, MeshEntities
from pymoab import types, rng
#mport pdb


class GetItem(object):
    def __init__(self, adj):
        self.fun = adj

    def __call__(self, item):
        return self.fun(item)

    def __getitem__(self, item):
        return self.fun(item)


class MeshEntitiesMS(MeshEntities):
    def __init__(self, core, entity_type):
        super().__init__(core, entity_type)
        self.father_handle = core.handleDic[core.father_id_name]
        self.global_id = GetItem(self._global_id)
        self.father_id = GetItem(self._father_id)

    def _global_id(self, index):
        range_vec = self.create_range_vec(index)
        element_handle = self.range_index(range_vec)
        return self.mb.tag_get_data(self.global_handle, element_handle).ravel()

    def _father_id(self, index):
        range_vec = self.create_range_vec(index)
        element_handle = self.range_index(range_vec)
        return self.mb.tag_get_data(self.father_handle, element_handle).ravel()

    def enhance(self,i, general):
        self._coarse_neighbors_dic = {}
        if self.vID == 0:
            index = general._nodes_neighbors[i, np.where(general.connectivities[i, :, 0])[0]].astype("uint64")
            self._coarse_neighbors_dic = {x: general._nodes[y] for x,y in zip(np.where(general.connectivities[i, :, 0])[0], index)}
        elif self.vID == 1:
            index = general._edges_neighbors[i, np.where(general.connectivities[i, :, 1])[0]].astype("uint64")
            self._coarse_neighbors_dic = {x: general._edges[y] for x,y in zip(np.where(general.connectivities[i, :, 1])[0], index)}
        elif self.vID == 2:
            index = general._faces_neighbors[i, np.where(general.connectivities[i, :, 2])[0]].astype("uint64")
            self._coarse_neighbors_dic = {x: general._faces[y] for x,y in zip(np.where(general.connectivities[i, :, 2])[0], index)}
        if self.vID < 3:
            self.coarse_neighbors = np.where(general.connectivities[i, :, self.vID])[0].astype("uint64")
            self.is_on_father_boundary = general.connectivities[i, -1, self.vID]
        self.neighborhood = GetItem(self._elements_in_coarse_neighborhood)

    def _elements_in_coarse_neighborhood(self,x):
        handle = self._coarse_neighbors_dic[x]
        return self.read(handle)

    @property
    def father_boundary(self):
        if self.is_on_father_boundary:
            handle = self.coarse_neighbors_dic[max(self.coarse_neighbors_dic.keys())]
            return self.read(handle)

    @property
    def all_coarse_neighbors(self):
        trange = rng.Range()
        for el in self.coarse_neighbors_dic.values():
            trange = rng.unite(trange, el)
        if self.is_on_father_boundary:
            trange = rng.subtract(trange, self.coarse_neighbors_dic[max(self.coarse_neighbors_dic.keys())])
        return self.read(trange)


class MoabVariableMS(MoabVariable):
    def __init__(self, core, name_tag, var_type="volumes", data_size=1, data_format="float", data_density="sparse",
                 entity_index=None, level=0, coarse_num=0):

        self.mb = core.mb
        self.var_type = var_type
        self.data_format = data_format
        self.data_size = data_size
        self.data_density = data_density
        self.name_tag = name_tag
        self.custom = False
        if var_type == "nodes":
            self.elements_handle = core.all_nodes
        elif var_type == "edges":
            self.elements_handle = core.all_edges
        elif var_type == "faces":
            self.elements_handle = core.all_faces
        elif var_type == "volumes":
            self.elements_handle = core.all_volumes
        if entity_index is not None:
            self.elements_handle = self.range_index(entity_index)
            self.custom = True
        if data_density == "dense":
            data_density = types.MB_TAG_DENSE
        elif data_density == "sparse":
            data_density = types.MB_TAG_SPARSE
        elif data_density == "bit":
            data_density = types.MB_TAG_BIT
        else:
            print("Please define a valid tag type")
        if data_format == 'float':
            data_format = types.MB_TYPE_DOUBLE
        elif data_format == "int":
            data_format = types.MB_TYPE_INTEGER
        elif data_format == "bool":
            data_format = types.MB_TYPE_BIT
        self.level = level
        self.coarse_num = coarse_num
        name = core.id_name
        name = name[(name.find("ID") + 3):]
        self.name_tag = self.name_tag  + name
        #import pdb; pdb.set_trace()
        #"-L" + str(self.level) + "-" + str(self.coarse_num)
        self.tag_handle = self.mb.tag_get_handle(self.name_tag, data_size, data_format, data_density, True)
        print("Component class {0} successfully intialized".format(self.name_tag))



"""
Module for management of fine scale mesh
"""
import time
import pdb
import numpy as np
from . corePymoab import CoreMoab
from . meshComponents import MeshEntities
import hdmedians as hd
import numpy as np
import scipy as sp
import scipy.spatial
import collections

print('Dual Coarse Mesh Module initialized')


class DualCoarseMesh:
    def __init__(self, M):
        self.M = M
        self.find_interface_centers()
        self.find_primal_coarse_centers()
        self.find_vol_neighbors_to_interface_center()
        # pdb.set_trace()

        self.find_coarse_edges()
        self.find_coarse_faces()
        # self.find_coarse_volumes()

    def find_primal_coarse_centers(self):
        self.coarse_center = np.zeros(len(self.M.coarse.elements)).astype("uint64")
        index = 0
        for index, coarse_volume in enumerate(self.M.coarse.elements):
            center_coord = np.array(hd.geomedian(coarse_volume.faces.center[coarse_volume.faces.boundary].T))
            center_coord = center_coord.reshape((1,3))
            distance = scipy.spatial.distance.cdist(coarse_volume.volumes.center[:],center_coord)
            tag = np.nonzero(distance == distance.min())[0]
            if len(tag) != 1:
                tag = tag[0]
            center_volume = coarse_volume.volumes.all[tag]
            self.coarse_center[index] = coarse_volume.volumes.father_id[center_volume]
            #index += 1

    def find_interface_centers(self):
        #self.center_face = []
        coords = self.M.faces.center[:]
        self.center_face = np.zeros((len(self.M.coarse.interfaces_faces), 1)).astype("uint64")
        index = 0
        for cface in self.M.coarse.interfaces_faces:
            tcoords = coords[cface].reshape(len(cface),3)
            center_coord = np.array(hd.geomedian(tcoords.T)).reshape((1,3))
            distance = scipy.spatial.distance.cdist(tcoords, center_coord)
            tag = np.nonzero(distance == distance.min())[0]
            if len(tag) != 1:
                tag = tag[0]
            self.center_face[index] = int(cface[tag].ravel())
            index += 1


    def find_vol_neighbors_to_interface_center(self):
        # internal faces
        internal_volumes = self.M.faces.bridge_adjacencies(self.center_face[0:self.M.coarse.num_internal_faces], interface ="faces", target = "volumes")
        external_volumes = self.M.faces.bridge_adjacencies(self.center_face[self.M.coarse.num_internal_faces:], interface ="faces", target = "volumes")
        self.interface_vol = np.vstack((internal_volumes,np.hstack((external_volumes,external_volumes))))
        tag = self.M.coarse.partition[:].ravel()[self.interface_vol][:,0] > self.M.coarse.partition[:].ravel()[self.interface_vol][:,1]
        #pdb.set_trace()
        self.interface_vol[tag,0], self.interface_vol[tag,1] = self.interface_vol[tag,1], self.interface_vol[tag,0]
        #pdb.set_trace()

    def find_coarse_edges(self):
        edges = np.array([])
        coarse_edges = [np.array([])]* len(self.M.coarse._faces)
        for x in range(len(self.M.coarse.elements)):
            cneigh, cfaces  = self.M.coarse.iface_neighbors(x)
            pointer = self.M.coarse.partition[:].ravel()[self.interface_vol[cfaces,:]] == x
            pointer_to_face = cneigh >= len(self.M.coarse.elements)
            pointer[pointer_to_face, :] = False
            target_volume = self.interface_vol[cfaces,:][pointer]
            target_volume_inside = self.interface_vol[cfaces,:][pointer_to_face,0]
            element_target = self.M.coarse.father_to_local_id(target_volume, "volumes", x).ravel()
            element_target_inside = self.M.coarse.father_to_local_id(target_volume_inside, "volumes", x).ravel()
            element_center = self.M.coarse.father_to_local_id(self.coarse_center[x], "volumes", x)
            shortest = GraphMesh(self.M.coarse.elements[x],center = element_center)
            # if x == 22:
            #     pdb.set_trace()
            for index, el in enumerate(element_target):

                # if index == 3 and x == 22:
                #     pdb.set_trace()
                #     print(index)
                center = self.M.coarse.elements[x].volumes.father_id[el]
                face = np.logical_or((self.interface_vol[:, 0] == center), (self.interface_vol[:, 1] == center))
                face = (np.where(face)[0])
                line = self.interface_vol[face,:]
                tag = np.zeros(line.shape[0],dtype=bool)
                #pdb.set_trace()
                tag = line == center
                for index, f in enumerate(face):
                    t = np.where(tag[index,:])[0]
                    m = self.M.coarse.elements[x].volumes.father_id[shortest.path(el)]
                    if t == 0:
                        coarse_edges[f] = np.append(m, coarse_edges[f])
                    else:
                        coarse_edges[f] = np.append(coarse_edges[f], m)

            for index, el in enumerate(element_target_inside):
                center = self.M.coarse.elements[x].volumes.father_id[el]
                face = np.logical_or((self.interface_vol[:, 0] == center), (self.interface_vol[:, 1] == center))
                face = int(np.where(face)[0])
                m = self.M.coarse.elements[x].volumes.father_id[shortest.path(el)]
                coarse_edges[face] = np.append(m, coarse_edges[face])
            #pdb.set_trace()
            self.coarse_edges = np.array(coarse_edges)

    def find_coarse_faces(self):
        pass
    def find_coarse_volumes(self):
        pass


class GraphMesh:
    def __init__(self, N, center):
        graph = self.create_sparse_matrix(N.faces.bridge_adjacencies(N.faces.internal, interface="faces", target="volumes"), len(N.volumes))
        # self.dist_matrix, self.predecessors = sp.sparse.csgraph.shortest_path(graph, directed=False, return_predecessors = True)
        self.dist_matrix, self.predecessors = sp.sparse.csgraph.dijkstra(graph, directed=False, return_predecessors = True, indices = center)
        self.predecessors = self.predecessors.ravel()

#         self.indicies = target
#         self.center = center
#         self.cedges = []
#         for el in target:
#             self.cedges.append(self.path(el).ravel())
# #             np.concatenate((self.cedges, self.path(el).ravel()))

    def create_sparse_matrix(self, graph_edges, size_vol):
        sparse_matrix = sp.sparse.coo_matrix((np.ones((len(graph_edges),),dtype=bool).T, (graph_edges[:, 0], graph_edges[:, 1])), shape=(size_vol, size_vol))
        return sparse_matrix

    def path(self, target):
        index = target
        path = []
        while index != -9999:
            path.append(index)
            index = self.predecessors[index]
        return np.array(path)
